import asyncio
import functools
import subprocess
import sys
import time

import pytest
from mamba_gator.handlers import ActionsStack


@pytest.mark.asyncio
async def test_ActionsStack_cancel():
    a = ActionsStack()
    dt = 0.01

    async def dummy_action():
        try:
            await asyncio.sleep(100.0 * dt)
        except asyncio.CancelledError:
            raise
        else:
            assert False, "CancelledError not raised"

        return True

    i = a.put(dummy_action)
    assert isinstance(i, int)

    await asyncio.sleep(dt)
    a.cancel(i)

    r = None
    elapsed = 0.0
    with pytest.raises(asyncio.CancelledError):
        while r is None and elapsed < 50 * dt:
            r = a.get(i)
            await asyncio.sleep(dt)  # Wait for task cancellation completion


@pytest.mark.asyncio
async def test_ActionsStack_cancel_subprocess():
    a = ActionsStack()
    dt = 0.01

    async def dummy_action():
        loop = asyncio.get_event_loop()
        code = f"import time; time.sleep(100.0 * {dt})"
        proc = await loop.run_in_executor(
            None,
            functools.partial(
                subprocess.Popen,
                [sys.executable, "-c", code],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ),
        )
        try:
            r = await loop.run_in_executor(None, proc.communicate)
            return r
        except asyncio.CancelledError:
            proc.terminate()
            await loop.run_in_executor(None, proc.wait)
            raise
        else:
            assert False, "CancelledError not raised"

    i = a.put(dummy_action)
    await asyncio.sleep(3.0 * dt)  # Give time to the processus to start
    a.cancel(i)

    r = None
    elapsed = 0.0
    with pytest.raises(asyncio.CancelledError):
        while r is None and elapsed < 50 * dt:
            r = a.get(i)
            await asyncio.sleep(dt)  # Wait for task cancellation completion


@pytest.mark.asyncio
async def test_ActionsStack_put_get():
    a = ActionsStack()
    dt = 0.01

    async def dummy_action():
        await asyncio.sleep(dt)
        return True

    i = a.put(dummy_action)
    assert isinstance(i, int)
    r = a.get(i)
    assert r is None

    elapsed = 0.0
    while r is None and elapsed < 50 * dt:
        elapsed += dt
        await asyncio.sleep(2 * dt)  # Wait for task completion
        r = a.get(i)
    assert r


@pytest.mark.asyncio
async def test_ActionsStack_put_result():
    a = ActionsStack()

    async def f(i):
        await asyncio.sleep(0.01)
        return i

    to_be_tested = [10, 20, 30]
    idxs = []

    for b in to_be_tested:
        idxs.append(a.put(f, b))

    len(idxs) == len(to_be_tested)
    for l in idxs:
        assert isinstance(l, int)
        assert a.get(l) is None

    for i, v in enumerate(to_be_tested):
        r = None
        elapsed = 0.0
        dt = 0.02
        while r is None and elapsed < 50 * dt:
            elapsed += dt
            await asyncio.sleep(dt)
            r = a.get(idxs[i])
        assert r == v



import logging
from typing import Optional

from traitlets.config import Application


class _ExtensionLogger:
    _LOGGER: Optional[logging.Logger] = None

    @classmethod
    def get_logger(cls) -> logging.Logger:
        if cls._LOGGER is None:
            app = Application.instance()
            cls._LOGGER = logging.getLogger("{!s}.mamba_gator".format(app.log.name))
            app.clear_instance()

        return cls._LOGGER


get_logger = _ExtensionLogger.get_logger



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import logging
import os

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.extension.handler import (
    ExtensionHandlerJinjaMixin,
    ExtensionHandlerMixin,
)
from jupyter_server.utils import url_path_join as ujoin
from jupyterlab_server import LabServerApp
from mamba_gator._version import __version__
from mamba_gator.handlers import _load_jupyter_server_extension
from mamba_gator.log import get_logger

HERE = os.path.dirname(__file__)


class MambaNavigatorHandler(
    ExtensionHandlerJinjaMixin, ExtensionHandlerMixin, JupyterHandler
):
    def get(self):
        config_data = {
            "appVersion": __version__,
            "baseUrl": self.base_url,
            "token": self.settings["token"],
            "fullStaticUrl": ujoin(self.base_url, "static", self.name),
            "frontendUrl": ujoin(self.base_url, "gator/"),
        }
        return self.write(
            self.render_template(
                "index.html",
                static=self.static_url,
                base_url=self.base_url,
                token=self.settings["token"],
                page_config=config_data,
            )
        )


class MambaNavigator(LabServerApp):
    name = __name__
    app_name = "Mamba Navigator"
    app_url = "/gator"
    default_url = "/gator"
    extension_url = "/gator"
    app_settings_dir = os.path.join(HERE, "application_settings")
    app_version = __version__
    load_other_extensions = False
    schemas_dir = os.path.join(HERE, "schemas")
    static_dir = os.path.join(HERE, "static")
    templates_dir = os.path.join(HERE, "templates")
    terminals = False
    themes_dir = os.path.join(HERE, "themes")
    user_settings_dir = os.path.join(HERE, "user_settings")
    workspaces_dir = os.path.join(HERE, "workspaces")

    def initialize_handlers(self):
        self.handlers.append(("/gator", MambaNavigatorHandler))
        super().initialize_handlers()

    def start(self):
        _load_jupyter_server_extension(self)
        super().start()


def main():
    MambaNavigator.launch_instance()


if __name__ == "__main__":
    main()



"""
# Copyright (c) 2015-2016 Continuum Analytics.
# Copyright (c) 2016-2020 Jupyter Development Team.
# See LICENSE.txt for the license.
"""
# pylint: disable=W0221

import asyncio
import collections
import json
import logging
import os
import re
import stat
import sys
import tempfile
import traceback
from typing import Any, Callable, ClassVar, Dict, NoReturn

import tornado

from .envmanager import EnvManager
from .log import get_logger
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join

NS = r"conda"
# Filename for the available conda packages list cache in temp folder
AVAILABLE_CACHE = "mamba_gator_packages"


class ActionsStack:
    """Process long asynchronous task.

    The task result can only be queried once.

    Note: The task start immediately => may bring trouble if the user is not careful,
    this assumes 'conda' handle lockers when appropriate.
    """

    __last_index: ClassVar[int] = 0

    def __init__(self):
        self.__tasks: Dict[int, asyncio.Task] = dict()

    def cancel(self, idx: int) -> NoReturn:
        """Cancel the task `idx`.

        Args:
            idx (int): Task index
        """
        get_logger().debug("Cancel task {}.".format(idx))
        if idx not in self.__tasks:
            raise ValueError("Task {} does not exists.".format(idx))

        self.__tasks[idx].cancel()

    def get(self, idx: int) -> Any:
        """Get the task `idx` results or None.

        Args:
            idx (int): Task index

        Returns:
            Any: None if the task is pending else its result

        Raises:
            ValueError: If the task `idx` does not exists.
            asyncio.CancelledError: If the task `idx` was cancelled.
        """
        if idx not in self.__tasks:
            raise ValueError("Task {} does not exists.".format(idx))

        if self.__tasks[idx].done():
            task = self.__tasks.pop(idx)
            return task.result()
        else:
            return None

    def put(self, task: Callable, *args) -> int:
        """Add a asynchronous task into the queue.

        Args:
            task (Callable): Asynchronous task
            *args : arguments of the task

        Returns:
            int: Task id
        """
        ActionsStack.__last_index += 1
        idx = ActionsStack.__last_index

        async def execute_task(idx, f, *args) -> Any:
            try:
                get_logger().debug("Will execute task {}.".format(idx))
                result = await f(*args)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                exception_type, _, tb = sys.exc_info()
                result = {
                    "type": exception_type.__qualname__,
                    "error": str(e),
                    "message": repr(e),
                    "traceback": traceback.format_tb(tb),
                }
                get_logger().error("Error for task {}.".format(result))
            else:
                get_logger().debug("Has executed task {}.".format(idx))

            return result

        self.__tasks[idx] = asyncio.ensure_future(execute_task(idx, task, *args))
        return idx

    def __del__(self):
        for task in filter(lambda t: not t.cancelled(), self.__tasks.values()):
            task.cancel()


class EnvBaseHandler(APIHandler):
    """
    Mixin for an env manager. Just maintains a reference to the
    'env_manager' which implements all of the conda functions.
    """

    _stack: ClassVar[ActionsStack] = ActionsStack()

    @property
    def env_manager(self) -> EnvManager:
        """EnvManager : Return our env_manager instance"""
        return self.settings["env_manager"]

    @property
    def log(self) -> logging.Logger:
        """logging.Logger: The extension logger"""
        return get_logger()

    def redirect_to_task(self, index: int):
        """Close a request by redirecting to a task."""
        self.set_status(202)
        self.set_header("Location", "/{}/tasks/{}".format(NS, index))
        self.finish("{}")


class ChannelsHandler(EnvBaseHandler):
    """Handle channels actions."""

    @tornado.web.authenticated
    async def get(self):
        """`GET /channels` list the channels."""
        channels = await self.env_manager.env_channels()
        if "error" in channels:
            self.set_status(500)
        self.finish(tornado.escape.json_encode(channels))


class EnvironmentsHandler(EnvBaseHandler):
    """Environments handler."""

    @tornado.web.authenticated
    async def get(self):
        """`GET /environments` which lists the environments.

        Query arguments:
            whitelist (int): optional flag 0 or 1 to respect KernelSpecManager.whitelist

        Raises:
            500 if an error occurs
        """
        whitelist = int(self.get_query_argument("whitelist", 0))
        list_envs = await self.env_manager.list_envs(whitelist == 1)
        if "error" in list_envs:
            self.set_status(500)
        self.finish(tornado.escape.json_encode(list_envs))

    @tornado.web.authenticated
    def post(self):
        """`POST /environments` creates an environment.

        Method of creation depends on the request data (first find is used):

        * packages: Create from a list of packages
        * twin: Clone the environment given by its name
        * file: Import from the given file content

        Request json body:
        {
            name (str): environment name
            packages (List[str]): optional, list of packages to install
            twin (str): optional, environment name to clone
            file (str): optional, environment file (TXT or YAML format)
            filename (str): optional, environment filename of the `file` content
        }
        """
        data = self.get_json_body()
        name = data["name"]
        packages = data.get("packages", None)
        twin = data.get("twin", None)
        file_content = data.get("file", None)
        file_name = data.get("filename", "environment.txt")

        if packages is not None:
            idx = self._stack.put(self.env_manager.create_env, name, *packages)
        elif twin is not None:
            idx = self._stack.put(self.env_manager.clone_env, twin, name)
        elif file_content is not None:
            idx = self._stack.put(
                self.env_manager.import_env, name, file_content, file_name
            )
        else:
            idx = self._stack.put(self.env_manager.create_env, name)

        self.redirect_to_task(idx)


class EnvironmentHandler(EnvBaseHandler):
    """Environment handler."""

    @tornado.web.authenticated
    def delete(self, env: str):
        """`DELETE /environments/<env>` deletes an environment."""
        idx = self._stack.put(self.env_manager.delete_env, env)

        self.redirect_to_task(idx)

    @tornado.web.authenticated
    async def get(self, env: str):
        """`GET /environments/<env>` List or export the environment packages.

        Query arguments:
            status: "installed" (default) or "has_update"
            download: 0 (default) or 1
            history: 0 (default) or 1
        """
        status = self.get_query_argument("status", "installed")
        download = int(self.get_query_argument("download", 0))
        history = int(self.get_query_argument("history", 0))

        if download:
            # export requirements file
            self.set_header(
                "Content-Disposition", 'attachment; filename="%s"' % (env + ".yml")
            )
            answer = await self.env_manager.export_env(env, bool(history))
            if "error" in answer:
                self.set_status(500)
                self.finish(tornado.escape.json_encode(answer))
            else:
                self.finish(answer)

        else:
            if status == "has_update":
                idx = self._stack.put(self.env_manager.check_update, env, ["--all"])

                self.redirect_to_task(idx)

            else:
                packages = await self.env_manager.env_packages(env)

                if "error" in packages:
                    self.set_status(500)
                self.finish(tornado.escape.json_encode(packages))

    @tornado.web.authenticated
    def patch(self, env: str):
        """`PATCH /environments/<env>` update an environment.

        Request json body:
        {
            file (str): optional, environment file (TXT or YAML format)
            filename (str): optional, environment filename of the `file` content
        }
        """
        data = self.get_json_body()
        file_content = data["file"]
        file_name = data.get("filename", "environment.yml")

        idx = self._stack.put(self.env_manager.update_env, env, file_content, file_name)

        self.redirect_to_task(idx)


class PackagesEnvironmentHandler(EnvBaseHandler):
    """Handle actions on environment packages."""

    @tornado.web.authenticated
    def delete(self, env: str):
        """`DELETE /environments/<env>/packages` delete some packages.

        Request json body:
        {
            packages (List[str]): list of packages to delete
        }
        """
        body = self.get_json_body()
        packages = body["packages"]
        idx = self._stack.put(self.env_manager.remove_packages, env, packages)
        self.redirect_to_task(idx)

    @tornado.web.authenticated
    def patch(self, env: str):
        """`PATCH /environments/<env>/packages` update some packages.

        If no packages are provided, update all possible packages.

        Request json body:
        {
            packages (List[str]): optional, list of packages to update
        }
        """
        body = self.get_json_body() or {}
        packages = body.get("packages", ["--all"])
        idx = self._stack.put(self.env_manager.update_packages, env, packages)
        self.redirect_to_task(idx)

    @tornado.web.authenticated
    def post(self, env: str):
        """`POST /environments/<env>/packages` install some packages.

        Packages can be installed in development mode. In that case,
        the fullpath to the package directory should be provided.

        Query arguments:
            develop: 0 (default) or 1

        Request json body:
        {
            packages (List[str]): list of packages to install
        }
        """
        body = self.get_json_body()
        packages = body["packages"]
        develop = int(self.get_query_argument("develop", 0))

        if develop:
            idx = self._stack.put(self.env_manager.develop_packages, env, packages)
        else:
            idx = self._stack.put(self.env_manager.install_packages, env, packages)
        self.redirect_to_task(idx)


class PackagesHandler(EnvBaseHandler):
    """Handles packages search"""

    __is_listing_available = False

    @tornado.web.authenticated
    async def get(self):
        """`GET /packages` Search for packages.

        Query arguments:
            dependencies: 0 (default) or 1
            query (str): optional string query
        """
        dependencies = self.get_query_argument("dependencies", 0)
        query = self.get_query_argument("query", "")

        idx = None
        if query:
            if dependencies:
                idx = self._stack.put(self.env_manager.pkg_depends, query)

            else:  # Specific search
                idx = self._stack.put(self.env_manager.package_search, query)

        else:  # List all available
            cache_file = os.path.join(tempfile.gettempdir(), AVAILABLE_CACHE + ".json")
            cache_data = ""
            try:
                with open(cache_file) as cache:
                    cache_data = cache.read()
            except OSError as e:
                self.log.info("No available packages list in cache.")
                self.log.debug(str(e))

            async def update_available(
                env_manager: EnvManager, cache_file: str, return_packages: bool = True
            ) -> Dict:
                answer = await env_manager.list_available()
                try:
                    with open(cache_file, "w+") as cache:
                        json.dump(answer, cache)
                except (ValueError, OSError) as e:
                    self.log.info("Fail to cache available packages.")
                    self.log.debug(str(e))
                else:
                    # Change rights to ensure every body can update the cache
                    os.chmod(
                        cache_file,
                        stat.S_IMODE(os.stat(cache_file).st_mode)
                        | (stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH),
                    )
                PackagesHandler.__is_listing_available = False

                if return_packages:
                    return answer
                else:
                    return {}

            if len(cache_data) > 0:
                self.log.debug("Loading available packages from cache.")
                # Request cache update in background
                if not PackagesHandler.__is_listing_available:
                    PackagesHandler.__is_listing_available = True
                    self._stack.put(
                        update_available, self.env_manager, cache_file, False
                    )
                # Return current cache
                self.set_status(200)
                self.finish(cache_data)
            else:
                # Request cache update and return once updated
                PackagesHandler.__is_listing_available = True
                idx = self._stack.put(update_available, self.env_manager, cache_file)

        if idx is not None:
            self.redirect_to_task(idx)


class TaskHandler(EnvBaseHandler):
    """Handler for /tasks/<id>"""

    @tornado.web.authenticated
    def get(self, index: int):
        """`GET /tasks/<id>` Returns the task `index` status.

        Status are:

        * 200: Task result is returned
        * 202: Task is pending
        * 500: Task ends with errors

        Args:
            index (int): Task index

        Raises:
            404 if task `index` does not exist
        """
        try:
            r = self._stack.get(int(index))
        except ValueError as err:
            raise tornado.web.HTTPError(404, reason=str(err))
        else:
            if r is None:
                self.set_status(202)
                self.finish("{}")
            else:
                if "error" in r:
                    self.set_status(500)
                    self.log.debug("{}".format(r))
                else:
                    self.set_status(200)
                self.finish(json.dumps(r))

    @tornado.web.authenticated
    def delete(self, index: int):
        """`DELETE /tasks/<id>` cancels the task `index`.

        Status are:
        * 204: Task cancelled

        Args:
            index (int): Task index

        Raises:
            404 if task `index` does not exist
        """
        try:
            self._stack.cancel(int(index))
        except ValueError as err:
            raise tornado.web.HTTPError(404, reason=str(err))
        else:
            self.set_status(204)
            self.finish()


# -----------------------------------------------------------------------------
# URL to handler mappings
# -----------------------------------------------------------------------------

# there is almost no text that is invalid, but no hyphens up front, please
# neither all these suspicious but valid caracthers...
_env_regex = r"(?P<env>[^/&+$?@<>%*-][^/&+$?@<>%*]*)"  # type: str


default_handlers = [
    (r"/channels", ChannelsHandler),
    (r"/environments", EnvironmentsHandler),  # GET / POST
    (r"/environments/%s" % _env_regex, EnvironmentHandler),  # GET / PATCH / DELETE
    # PATCH / POST / DELETE
    (r"/environments/%s/packages" % _env_regex, PackagesEnvironmentHandler),
    (r"/packages", PackagesHandler),  # GET
    (r"/tasks/%s" % r"(?P<index>\d+)", TaskHandler),  # GET / DELETE
]


def _load_jupyter_server_extension(server_app):
    """Load the nbserver extension"""
    webapp = server_app.web_app
    webapp.settings["env_manager"] = EnvManager(
        server_app.contents_manager.root_dir, server_app.kernel_spec_manager
    )

    base_url = webapp.settings["base_url"]
    webapp.add_handlers(
        ".*$",
        [
            (url_path_join(base_url, NS, pat), handler)
            for pat, handler in default_handlers
        ],
    )

    get_logger().info("Server extension enabled")



# Copyright (c) 2016-2020 Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import asyncio
import collections
import json
import logging
import os
import re
import sys
import tempfile
from functools import lru_cache, partial
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, Dict, List, Optional, Tuple, Union

import tornado
from jupyter_client.kernelspec import KernelSpecManager
from packaging.version import InvalidVersion, Version, parse

try:
    import nb_conda_kernels
except ImportError:
    nb_conda_kernels = None

from jupyter_server.utils import url2path, url_path_join

from .log import get_logger

CONDA_EXE = os.environ.get("CONDA_EXE", "conda")  # type: str

PATH_SEP = "\\" + os.path.sep
CONDA_ENV_PATH = r"^(.*?" + PATH_SEP + r"envs" + PATH_SEP + r".+?)" + PATH_SEP

# try to match lines of json
JSONISH_RE = r'(^\s*["\{\}\[\],\d])|(["\}\}\[\],\d]\s*$)'  # type: str

MAX_LOG_OUTPUT = 6000  # type: int

ROOT_ENV_NAME = "base"

# See https://github.com/Anaconda-Platform/nb_conda_kernels/blob/master/nb_conda_kernels/manager.py#L19
try:
    RUNNER_COMMAND = nb_conda_kernels.manager.RUNNER_COMMAND
except AttributeError:
    RUNNER_COMMAND = ["python", "-m", "nb_conda_kernels.runner"]


def normalize_pkg_info(s: Dict[str, Any]) -> Dict[str, Union[str, List[str]]]:
    """Normalize package information.

    Args:
        s (dict): Raw package information

    Returns:
        dict: Normalized package information
    """
    return {
        "build_number": s.get("build_number"),
        "build_string": s.get("build_string", s.get("build")),
        "channel": s.get("channel"),
        "name": s.get("name"),
        "platform": s.get("platform"),
        "version": s.get("version"),
        "summary": s.get("summary", ""),
        "home": s.get("home", ""),
        "keywords": s.get("keywords", []),
        "tags": s.get("tags", []),
    }


def get_env_path(kernel_spec: Dict[str, Any]) -> Optional[str]:
    """Get the conda environment path.

    Args:
        kernel_spec (dict): Kernel spec

    Returns:
        str: Best guess for the environment path
    """
    argv = kernel_spec.get("argv", [])
    if "conda_env_path" in kernel_spec["metadata"]:
        return kernel_spec["metadata"]["conda_env_path"]
    elif len(argv) >= 5 and argv[:3] == RUNNER_COMMAND and len(argv[4]) > 0:
        return argv[4]
    elif len(argv) > 0:
        match = re.match(CONDA_ENV_PATH, argv[0])
        if match is not None:
            return match.groups()[0]

    return None


def parse_version(version: str) -> Optional[Version]:
    """Handle R-style and year-based versions"""
    # Convert R package versions like "1.8_4" to "1.8.4"
    version = version.replace('_', '.')

    # Handle year-based versions like "2023d" -> "2023.4"
    if re.match(r'^\d{4}[a-z]$', version):
        letter = version[-1]
        number = ord(letter) - ord('a') + 1
        version = f"{version[:-1]}.{number}"

    try:
        return Version(version)
    except InvalidVersion:
        return None


class EnvManager:
    """Handles environment and package actions."""

    _conda_version: Optional[str] = None
    _mamba_version: Optional[str] = None
    _manager_exe: Optional[str] = None

    def __init__(self, root_dir: str, kernel_spec_manager: KernelSpecManager):
        """
        Args:
            root_dir (str): Server root path
        """
        self._root_dir = root_dir
        self._kernel_spec_manager = kernel_spec_manager

    def _clean_conda_json(self, output: str) -> Dict[str, Any]:
        """Clean a command output to fit json format.

        Args:
            output (str): output to clean

        Returns:
            Dict[str, Any]: Cleaned output
        """
        lines = output.splitlines()

        if not output.strip():
            return {}

        try:
            return json.loads("\n".join(lines))
        except (ValueError, json.JSONDecodeError) as err:
            self.log.warning("JSON parse fail:\n{!s}".format(err))

        # try to remove bad lines
        lines = [line for line in lines if re.match(JSONISH_RE, line)]

        try:
            return json.loads("\n".join(lines))
        except (ValueError, json.JSONDecodeError) as err:
            self.log.error("JSON clean/parse fail:\n{!s}".format(err))

        return {"error": True}

    async def _execute(self, cmd: str, *args) -> Tuple[int, str]:
        """Asynchronously execute a command.

        Args:
            cmd (str): command to execute
            *args: additional command arguments

        Returns:
            (int, str): (return code, output) or (return code, error)
        """
        cmdline = [cmd]
        cmdline.extend(args)

        self.log.debug("command: {!s}".format(" ".join(cmdline)))

        current_loop = tornado.ioloop.IOLoop.current()
        
        # Set environment variables to suppress Windows file association dialogs
        env = os.environ.copy()
        subprocess_kwargs = {"stdout": PIPE, "stderr": PIPE, "env": env}
        
        if sys.platform == "win32":
            env["PATHEXT"] = env.get("PATHEXT", "") + ";.env"  # Treat .env as executable to avoid dialog
        
        process = await current_loop.run_in_executor(
            None, partial(Popen, cmdline, **subprocess_kwargs)
        )
        try:
            output, error = await current_loop.run_in_executor(
                None, process.communicate
            )
        except asyncio.CancelledError:
            process.terminate()
            await current_loop.run_in_executor(None, process.wait)
            raise

        returncode = process.returncode
        if returncode == 0:
            output = output.decode("utf-8")
        else:
            self.log.debug("exit code: {!s}".format(returncode))
            output = error.decode("utf-8") + output.decode("utf-8")

        self.log.debug("output: {!s}".format(output[:MAX_LOG_OUTPUT]))

        if len(output) > MAX_LOG_OUTPUT:
            self.log.debug("...")

        return returncode, output

    @property
    def log(self) -> logging.Logger:
        """logging.Logger : Extension logger"""
        return get_logger()

    @property
    def manager(self) -> str:
        """Conda package manager name.

        For now, use mamba if it is installed. Otherwise, fallback to conda.

        Returns:
            str: Package manager
        """
        if EnvManager._manager_exe is None:
            # Set conda by default
            EnvManager._manager_exe = CONDA_EXE
            try:
                cmd = ["which", "mamba"]
                if sys.platform == "win32":
                    cmd = ["where", "mamba.exe"]

                process = Popen(
                    cmd, stdout=PIPE, stderr=PIPE, encoding="utf-8"
                )
                output, error = process.communicate()

                if process.returncode != 0:
                    raise RuntimeError(error)

                mamba_exe = output.splitlines()[0] or "mamba"

                process = Popen(
                    [mamba_exe, "--version"],
                    stdout=PIPE,
                    stderr=PIPE,
                    encoding="utf-8",
                )
                output, error = process.communicate()

                if process.returncode != 0:
                    raise RuntimeError(error)

                versions = list(map(lambda l: l.split(), output.splitlines()))
                if versions[0][0] == "mamba" and versions[1][0] == "conda":
                    EnvManager._conda_version = versions[1][1]
                    EnvManager._mamba_version = versions[0][1]
                    EnvManager._manager_exe = mamba_exe

            except BaseException:
                self.log.debug(
                    "Fail to get mamba version, falling back to conda",
                    exc_info=sys.exc_info(),
                )

            self.log.debug("Package manager: {}".format(EnvManager._manager_exe))

        return EnvManager._manager_exe

    @lru_cache()
    def is_mamba(self):
        return Path(self.manager).stem == "mamba"

    async def env_channels(
        self, configuration: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Dict[str, List[str]]]:
        """List available channels.

        Args:
            configuration (Dict[str, Any] or None): Conda configuration

        Returns:
            {"channels": {<channel>: <uri>}}
        """
        if configuration is None:
            info = await self.conda_config()
        else:
            info = configuration

        if "error" in info:
            return info

        deployed_channels = {}

        def get_uri(spec):
            location = "/".join((spec["location"], spec["name"]))
            return spec["scheme"] + "://" + location

        for channel in info["channels"]:
            strip_channel = channel.strip("/")
            if channel in info["custom_multichannels"]:
                deployed_channels[channel] = [
                    get_uri(entry) for entry in info["custom_multichannels"][channel]
                ]
            elif strip_channel in info["custom_channels"]:
                deployed_channels[strip_channel] = [
                    get_uri(info["custom_channels"][strip_channel])
                ]
            else:
                parsed_channel = tornado.httputil.urlparse(channel)
                if parsed_channel.scheme:
                    deployed_channels[strip_channel] = [
                        channel.strip("/"),
                    ]
                else:
                    spec = info["channel_alias"]
                    spec["name"] = channel
                    deployed_channels[channel] = [
                        get_uri(spec),
                    ]

        self.log.debug("channels: {}".format(deployed_channels))
        return {"channels": deployed_channels}

    async def conda_config(self) -> Dict[str, Any]:
        """Get conda configuration.

        Returns:
            Dict[str, Any]: Conda configuration
        """
        ans = await self._execute(CONDA_EXE, "config", "--show", "--json")
        _, output = ans
        return self._clean_conda_json(output)

    async def clone_env(self, env: str, name: str) -> Dict[str, str]:
        """Clone an environment.

        Args:
            env (str): To-be-cloned environment name
            name (str): New environment name

        Returns:
            Dict[str, str]: Clone command output.
        """
        ans = await self._execute(
            self.manager, "create", "-y", "-q", "--json", "-n", name, "--clone", env
        )

        rcode, output = ans
        if rcode > 0:
            return {"error": output}
        
        return self._clean_conda_json(output)

    async def create_env(self, env: str, *args) -> Dict[str, str]:
        """Create a environment from a list of packages.

        Args:
            env (str): Name of the environment
            *args (List[str]): optional, packages to install

        Returns:
            Dict[str, str]: Create command output
        """
        ans = await self._execute(
            self.manager, "create", "-y", "-q", "--json", "-n", env, *args
        )

        rcode, output = ans
        if rcode > 0:
            return {"error": output}
        
        return self._clean_conda_json(output)

    async def delete_env(self, env: str) -> Dict[str, str]:
        """Delete an environment.

        Args:
            env (str): Environment name

        Returns:
            Dict[str, str]: Deletion command output
        """
        ans = await self._execute(
            self.manager, "env", "remove", "-y", "-q", "--json", "-n", env
        )

        rcode, output = ans
        if rcode > 0:
            # Check if this is just a "environment not found" error, which should be treated as success
            try:
                error_data = self._clean_conda_json(output)
                if isinstance(error_data, dict) and error_data.get("exception_name") == "EnvironmentLocationNotFound":
                    # Environment doesn't exist, treat as successful deletion
                    return {"success": True, "message": f"Environment '{env}' was already removed or did not exist"}
            except (json.JSONDecodeError, KeyError):
                pass
            
            return {"error": output}

        return self._clean_conda_json(output)

    async def export_env(
        self, env: str, from_history: bool = False
    ) -> Union[str, Dict[str, str]]:
        """Export an environment as YAML file.

        Args:
            env (str): Environment name
            from_history (bool): If True, use `--from-history` option; default False

        Returns:
            str: YAML file content
        """
        command = [self.manager, "env", "export", "-n", env]
        if from_history:
            if EnvManager._conda_version is None:
                await self.info()  # Set conda version
            if EnvManager._conda_version < (4, 7, 12):
                self.log.warning(
                    "conda<4.7.12 does not support `env export --from-history`. It will be ignored."
                )
            else:
                command.append("--from-history")
        ans = await self._execute(*command)
        rcode, output = ans
        if rcode > 0:
            return {"error": output}
        return output

    async def import_env(
        self, env: str, file_content: str, file_name: str = "environment.txt"
    ) -> Dict[str, str]:
        """Create an environment from a file.

        Args:
            env (str): Environment name
            file_content (str): File content
            file_name (str): optional, Original filename

        Returns:
            Dict[str, str]: Create command output
        """
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=file_name) as f:
            name = f.name
            f.write(file_content)

        ans = await self._execute(
            self.manager, "env", "create", "-y", "-q", "--json", "-n", env, "--file", name
        )

        # Remove temporary file
        os.unlink(name)

        rcode, output = ans
        if rcode > 0:
            return {"error": output}
        
        return self._clean_conda_json(output)

    async def info(self) -> Dict[str, Any]:
        """Returns `conda info --json` execution.

        Returns:
            The dictionary of conda information
        """
        ans = await self._execute(self.manager, "info", "--json")
        rcode, output = ans
        
        if rcode != 0:
            error_data = self._clean_conda_json(output)
            if isinstance(error_data, dict) and "message" in error_data:
                return {"error": True, "message": error_data["message"]}
            return {"error": True, "message": output}
            
        info = self._clean_conda_json(output)
        
        conda_version = info.get("conda_version")
        if conda_version is not None:
            EnvManager._conda_version = tuple(
                map(
                    lambda part: int(part),
                    info.get("conda_version", EnvManager._conda_version).split("."),
                )
            )
        return info

    async def list_envs(
        self, whitelist: bool = False
    ) -> Dict[str, List[Dict[str, Union[str, bool]]]]:
        """List all environments that conda knows about.

        Args:
            whitelist (bool): optional, filter the environment list to respect
                KernelSpecManager.whitelist (default: False)

        An environment is described by a dictionary:
        {
            name (str): environment name,
            dir (str): environment prefix,
            is_default (bool): is the root environment
        }

        Returns:
            {"environments": List[env]}: The environments
        """
        info = await self.info()
        if "error" in info:
            return info

        default_env = info["default_prefix"]

        root_env = {
            "name": ROOT_ENV_NAME,
            "dir": info["root_prefix"],
            "is_default": info["root_prefix"] == default_env,
        }

        whitelist_env = set()
        if whitelist:
            # Build env path list - simplest way to compare kernel and environment
            for entry in self._kernel_spec_manager.get_all_specs().values():
                path = get_env_path(entry["spec"])
                if path:
                    whitelist_env.add(path)
            
            # Fallback: match environment names against whitelist patterns
            if not whitelist_env and self._kernel_spec_manager.whitelist:
                all_env_dirs = [info["root_prefix"]] + info.get("envs", [])
                for env_dir in all_env_dirs:
                    env_name = os.path.basename(env_dir) if env_dir != info["root_prefix"] else "base"
                    if f"conda-env-{env_name}-py" in self._kernel_spec_manager.whitelist:
                        whitelist_env.add(env_dir)

        def get_info(env):
            base_dir = os.path.dirname(env)
            if base_dir not in info["envs_dirs"]:
                return None

            return {
                "name": os.path.basename(env),
                "dir": env,
                "is_default": env == default_env,
            }

        envs_list = []
        
        # Add root environment if not whitelisting, or if whitelisting found valid matches, or if root is explicitly whitelisted
        if not whitelist or whitelist_env or root_env["dir"] in whitelist_env:
            envs_list.append(root_env)
            
        for env in info["envs"]:
            env_info = get_info(env)
            if env_info is not None and (not whitelist or env_info["dir"] in whitelist_env):
                envs_list.append(env_info)

        return {"environments": envs_list}

    async def update_env(
        self, env: str, file_content: str, file_name: str = "environment.yml"
    ) -> Dict[str, str]:
        """Update a environment from a file.

        Args:
            env (str): Name of the environment
            file_content (str): File content
            file_name (str): optional, Original filename

        Returns:
            Dict[str, str]: Update command output
        """
        # Get just the file extension for the temporary file
        ext = Path(file_name).suffix
        if not ext:
            ext = '.yml'  # Default to .yml if no extension

        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=ext) as f:
            name = f.name
            f.write(file_content)

        # Determine the command based on file type
        if file_name.endswith('.txt'):
            # For .txt files (explicit package lists), use conda install
            self.log.debug(f"Updating environment {env} with txt file using conda install")
            ans = await self._execute(
                self.manager, "install", "-y", "-q", "--json", "-n", env, "--file", name
            )
        else:
            # For .yml files (environment definitions), use conda env update
            self.log.debug(f"Updating environment {env} with yml file using conda env update")
            ans = await self._execute(
                self.manager, "env", "update", "-q", "--json", "-n", env, "-f", name
            )

        # Remove temporary file
        os.unlink(name)

        rcode, output = ans

        if rcode > 0:
            self.log.error(f"Error updating environment {env}: {output}")
            return {"error": output}

        return self._clean_conda_json(output)

    async def env_packages(self, env: str) -> Dict[str, List[str]]:
        """List environment package.

        Args:
            env (str): Environment name

        Returns:
            {"packages": List[package]}
        """
        ans = await self._execute(self.manager, "list", "--json", "-n", env)
        _, output = ans
        data = self._clean_conda_json(output)

        # Data structure
        #   List of dictionary. Example:
        # {
        #     "base_url": null,
        #     "build_number": 0,
        #     "build_string": "py36_0",
        #     "channel": "defaults",
        #     "dist_name": "anaconda-client-1.6.14-py36_0",
        #     "name": "anaconda-client",
        #     "platform": null,
        #     "version": "1.6.14"
        # }

        if "error" in data:
            # we didn't get back a list of packages, we got a dictionary with
            # error info
            return data

        # Process packages and identify development mode packages
        packages = []
        for package in data:
            normalized_pkg = normalize_pkg_info(package)
            
            # Check if this is a pip package installed in development mode
            # Development mode packages have 'channel': '<develop>' in pip list output
            # or are pip packages with editable installs
            if normalized_pkg.get("channel") == "<develop>":
                # Already marked as development mode by conda/pip
                pass
            elif normalized_pkg.get("channel") == "pypi":
                # This is a pip package - check if it's in development mode
                # We need to check if it's installed as editable
                envs = await self.list_envs()
                env_info = None
                for e in envs["environments"]:
                    if e["name"] == env:
                        env_info = e
                        break
                
                if env_info:
                    if sys.platform == "win32":
                        python_cmd = os.path.join(env_info["dir"], "python")
                    else:
                        python_cmd = os.path.join(env_info["dir"], "bin", "python")
                    
                    # Check if this package is installed in editable mode
                    try:
                        ans = await self._execute(
                            python_cmd, "-m", "pip", "list", "--editable", "--format=json"
                        )
                        pip_rcode, pip_output = ans
                        if pip_rcode == 0:
                            editable_packages = json.loads(pip_output)
                            for editable_pkg in editable_packages:
                                if editable_pkg["name"].lower() == normalized_pkg["name"].lower():
                                    normalized_pkg["channel"] = "<develop>"
                                    break
                    except (json.JSONDecodeError, Exception):
                        # If we can't determine editable status, leave as is
                        pass
            
            packages.append(normalized_pkg)

        return {"packages": packages}

    async def pkg_depends(self, pkg: str) -> Dict[str, List[str]]:
        """List environment packages dependencies.

        Args:
            pkg (str): Package name

        Returns:
            {"package": List[dependencies]}
        """
        if not self.is_mamba():
            self.log.warning(
                "Package manager '{}' does not support dependency query.".format(
                    self.manager
                )
            )
            return {pkg: None}

        resp = {}
        ans = await self._execute(self.manager, "repoquery", "depends", "--json", pkg)
        _, output = ans
        query = self._clean_conda_json(output)

        if "error" not in query:
            for dep in query["result"]["pkgs"]:
                if isinstance(dep, dict):
                    deps = dep.get("depends", None)
                    if deps:
                        resp[dep["name"]] = deps
                    else:
                        resp[dep["name"]] = []

        return resp


    async def list_available(self) -> Dict[str, List[Dict[str, str]]]:
        """List all available packages

        Returns:
            {
                "packages": List[package],
                "with_description": bool  # Whether we succeed in get some channeldata.json files
            }
        """
        if self.is_mamba():
            ans = await self._execute(self.manager, "repoquery", "search", "*", "--json")
        else:
            ans = await self._execute(self.manager, "search", "--json")
        _, output = ans

        current_loop = tornado.ioloop.IOLoop.current()
        data = await current_loop.run_in_executor(None, self._clean_conda_json, output)

        if "error" in data:
            # we didn't get back a list of packages, we got a
            # dictionary with error info
            return data

        def process_mamba_repoquery_output(data: Dict) -> Dict:
            """Make a dictionary with keys as packages name and values
            containing the list of available packages to match the json output
            of "conda search --json".
            """

            data_ = collections.defaultdict(lambda: [])
            for entry in data["result"]["pkgs"]:
                name = entry.get("name")
                if name is not None:
                    data_[name].append(entry)

            return data_

        if self.is_mamba():
            data = await current_loop.run_in_executor(None, process_mamba_repoquery_output, data)

        def format_packages(data: Dict) -> List:
            packages = []

            # Data structure
            #  Dictionary with package name key and value is a list of dictionary. Example:
            #  {
            #   "arch": "x86_64",
            #   "build": "np17py33_0",
            #   "build_number": 0,
            #   "channel": "https://repo.anaconda.com/pkgs/free/win-64",
            #   "constrains": [],
            #   "date": "2013-02-20",
            #   "depends": [
            #     "numpy 1.7*",
            #     "python 3.3*"
            #   ],
            #   "fn": "astropy-0.2-np17py33_0.tar.bz2",
            #   "license": "BSD",
            #   "md5": "3522090a8922faebac78558fbde9b492",
            #   "name": "astropy",
            #   "platform": "win32",
            #   "size": 3352442,
            #   "subdir": "win-64",
            #   "url": "https://repo.anaconda.com/pkgs/free/win-64/astropy-0.2-np17py33_0.tar.bz2",
            #   "version": "0.2"
            # }

            # List all available version for packages
            for entries in data.values():
                pkg_entry = None
                versions = list()
                max_build_numbers = list()
                max_build_strings = list()

                for entry in entries:
                    entry = normalize_pkg_info(entry)
                    if pkg_entry is None:
                        pkg_entry = entry
                    version = parse_version(entry.get("version", ""))

                    if version is None:
                        name = entry.get("name")
                        original_version = entry.get("version")
                        self.log.warning(f"Unable to parse version '{original_version}' of '{name}'")
                        version = Version("0.0.0")

                    if version not in versions:
                        versions.append(version)
                        max_build_numbers.append(entry.get("build_number", 0))
                        max_build_strings.append(entry.get("build_string", ""))
                    else:
                        version_idx = versions.index(version)
                        build_number = entry.get("build_number", 0)
                        if build_number > max_build_numbers[version_idx]:
                            max_build_numbers[version_idx] = build_number
                            max_build_strings[version_idx] = entry.get("build_string", "")

                sorted_versions_idx = sorted(range(len(versions)), key=versions.__getitem__)

                pkg_entry["version"] = [str(versions[i]) for i in sorted_versions_idx]
                pkg_entry["build_number"] = [max_build_numbers[i] for i in sorted_versions_idx]
                pkg_entry["build_string"] = [max_build_strings[i] for i in sorted_versions_idx]

                packages.append(pkg_entry)
            return packages

        packages = await current_loop.run_in_executor(None, format_packages, data)

        # Get channel short names
        configuration = await self.conda_config()
        channels = await self.env_channels(configuration)
        channels = channels["channels"]
        tr_channels = {}
        for short_name, channel in channels.items():
            tr_channels.update({uri: short_name for uri in channel})

        # # Get top channel URI to request channeldata.json
        # top_channels = set()
        # for uri in tr_channels:
        #     channel, arch = os.path.split(uri)
        #     if arch in PLATFORMS:
        #         top_channels.add(channel)
        #     else:
        #         top_channels.add(uri)

        # Request channeldata.json
        pkg_info = {}
        client = tornado.httpclient.AsyncHTTPClient(force_instance=True)
        for channel in tr_channels:
            url = tornado.httputil.urlparse(channel)
            if url.scheme == "file":
                if url.netloc:
                    path = "".join(("//", url.netloc, url.path))
                elif sys.platform == "win32":
                    path = url.path.lstrip("/")
                else:
                    path = url.path
                path = os.path.join(path, "channeldata.json")
                self.log.debug("Reading {}".format(path))
                try:  # Skip if file is not accessible
                    with open(path) as f:
                        channeldata = json.load(f)
                except (json.JSONDecodeError, OSError, ValueError) as err:
                    self.log.info("{!s} skipped".format(path))
                    self.log.debug(str(err))
                else:
                    pkg_info.update(channeldata["packages"])
            else:
                try:  # Skip if file is not accessible
                    response = await client.fetch(
                        tornado.httpclient.HTTPRequest(
                            url_path_join(channel, "channeldata.json"),
                            headers={"Content-Type": "application/json"},
                            validate_cert=configuration.get("ssl_verify", True),
                        )
                    )
                except Exception as e:
                    self.log.info("{}/channeldata.json skipped.".format(channel))
                    self.log.debug(str(e))
                else:
                    channeldata = response.body.decode("utf-8")
                    try:
                        pkg_info.update(json.loads(channeldata)["packages"])
                    except (json.JSONDecodeError, ValueError) as error:
                        self.log.info("{}/channeldata.json skipped.".format(channel))
                        self.log.debug(str(error))

        # Example structure channeldata['packages'] for channeldata_version == 1
        # "tmpc0d7d950": {
        #     "activate.d": false,
        #     "binary_prefix": false,
        #     "deactivate.d": false,
        #     "identifiers": [],
        #     "keywords": [
        #         "['package', 'tmpc0d7d950']"
        #     ],
        #     "license": "MIT",
        #     "post_link": false,
        #     "pre_link": false,
        #     "pre_unlink": false,
        #     "reference_package": "win-64/tmpc0d7d950-0.1.0.dev1-py36_0.tar.bz2",
        #     "run_exports": {},
        #     "subdirs": [
        #         "win-64"
        #     ],
        #     "summary": "Dummy package",
        #     "tags": [],
        #     "text_prefix": false,
        #     "version": "0.1.0.dev1"
        # }

        def update_packages(packages, pkg_info, tr_channels):
            # Update channel and add some info
            for package in packages:
                name = package["name"]
                if name in pkg_info:
                    package["summary"] = pkg_info[name].get("summary", "")
                    package["home"] = pkg_info[name].get("home", "")
                    # May return None so "or" with empty list
                    package["keywords"] = pkg_info[name].get("keywords", []) or []
                    package["tags"] = pkg_info[name].get("tags", []) or []

                # Convert to short channel names
                channel, _ = os.path.split(package["channel"])
                if channel in tr_channels:
                    package["channel"] = tr_channels[channel]

            return sorted(packages, key=lambda entry: entry.get("name"))

        packages = await current_loop.run_in_executor(
            None, update_packages, packages, pkg_info, tr_channels
        )

        return {
            "packages": packages,
            "with_description": len(pkg_info) > 0,
        }

    async def package_search(self, q: str) -> Dict[str, List]:
        """Search packages.

        Args:
            q (str): Search query

        Returns:
            {
                "packages": List[package],
                "with_description": bool  # Whether we succeed in get some channeldata.json files
            }
        """
        ans = await self._execute(self.manager, "search", "--json", q)
        _, output = ans
        data = self._clean_conda_json(output)

        if "error" in data:
            # we didn't get back a list of packages, we got a dictionary with
            # error info
            return data

        packages = []

        for entries in data.values():
            max_version = None
            max_version_entry = None

            for entry in entries:
                version = parse(entry.get("version", ""))

                if max_version is None or version > max_version:
                    max_version = version
                    max_version_entry = entry

            packages.append(max_version_entry)

        return {
            "packages": sorted(packages, key=lambda entry: entry.get("name")),
            "with_description": False,
        }

    async def check_update(
        self, env: str, packages: List[str]
    ) -> Dict[str, List[Dict[str, str]]]:
        """Check for packages update in an environment.

        if '--all' is the only element in `packages`, search for all
        possible update in the given environment.

        Args:
            env (str): Environment name
            packages (List[str]): List of packages to search for update

        Returns:
            {"updates": List[package]}
        """
        ans = await self._execute(
            self.manager, "update", "--dry-run", "-q", "--json", "-n", env, *packages
        )
        _, output = ans
        data = self._clean_conda_json(output)

        # Data structure in LINK
        #   List of dictionary. Example:
        # {
        #     "base_url": null,
        #     "build_number": 0,
        #     "build_string": "mkl",
        #     "channel": "defaults",
        #     "dist_name": "blas-1.0-mkl",
        #     "name": "blas",
        #     "platform": null,
        #     "version": "1.0"
        # }

        if "error" in data:
            # we didn't get back a list of packages, we got a dictionary with
            # error info
            return data
        elif "actions" in data:
            links = data["actions"].get("LINK", [])
            package_versions = [link for link in links]
            return {
                "updates": [
                    normalize_pkg_info(pkg_version) for pkg_version in package_versions
                ]
            }
        else:
            # no action plan returned means everything is already up to date
            return {"updates": []}

    async def install_packages(self, env: str, packages: List[str]) -> Dict[str, str]:
        """Install packages in an environment.

        Args:
            env (str): Environment name
            packages (List[str]): List of packages to install

        Returns:
            Dict[str, str]: Install command output.
        """
        ans = await self._execute(
            self.manager, "install", "-y", "-q", "--json", "-n", env, *packages
        )
        _, output = ans
        return self._clean_conda_json(output)

    async def develop_packages(
        self, env: str, packages: List[str]
    ) -> Dict[str, List[Dict[str, str]]]:
        """Install packages in pip editable mode in an environment.

        Args:
            env (str): Environment name
            packages (List[str]): List of packages to install

        Returns:
            Dict[str, str]: Install command output.
        """
        envs = await self.list_envs()
        if "error" in envs:
            return envs

        env_rootpath = None
        for e in envs["environments"]:
            if e["name"] == env:
                env_rootpath = e
                break

        if not env_rootpath:
            return {"error": f"Environment {env} not found"}

        # Ensure pip is installed in the environment
        pip_check = await self._execute(self.manager, "list", "-n", env, "pip", "--json")
        pip_rcode, pip_output = pip_check
        
        if pip_rcode != 0 or not pip_output.strip() or pip_output.strip() == "[]":
            pip_install = await self.install_packages(env, ["pip"])
            if "error" in pip_install:
                return {"error": f"Failed to install pip in environment {env}: {pip_install['error']}"}

        result = []
        
        # Get the Python executable path using pathlib
        env_path = Path(env_rootpath["dir"])
        if sys.platform == "win32":
            python_cmd = env_path / "python.exe"
        else:
            python_cmd = env_path / "bin" / "python"

        if not python_cmd.exists():
            return {"error": f"Python executable not found at {python_cmd}"}

        root_dir = Path(self._root_dir)
        
        for path in packages:
            # Try as a local filesystem path first
            pkg_path = Path(os.path.expanduser(path))
            if not pkg_path.is_absolute():
                pkg_path = pkg_path.resolve()
                
            if not pkg_path.exists():
                # Try as a JupyterLab path
                jupyterlab_path = url2path(path)
                pkg_path = root_dir / jupyterlab_path
                
                if not pkg_path.exists():
                    return {"error": f"Unable to find package path: {path}"}

            # Verify the package structure
            if not ((pkg_path / "setup.py").exists() or (pkg_path / "pyproject.toml").exists()):
                return {"error": f"Invalid package structure at {path}. Missing setup.py or pyproject.toml"}

            # Run pip install in editable mode
            ans = await self._execute(
                str(python_cmd),
                "-m",
                "pip",
                "install",
                "--progress-bar",
                "off",
                "-e",
                str(pkg_path),
            )
            rcode, output = ans
            
            if rcode > 0:
                return {"error": f"Failed to install {path} in development mode: {output}"}
            
            result.append({"path": str(pkg_path), "output": output})

        return {"packages": result}

    async def update_packages(self, env: str, packages: List[str]) -> Dict[str, str]:
        """Update packages in an environment.

        Args:
            env (str): Environment name
            packages (List[str]): List of packages to update

        Returns:
            Dict[str, str]: Update command output.
        """
        ans = await self._execute(
            self.manager, "update", "-y", "-q", "--json", "-n", env, *packages
        )
        _, output = ans
        return self._clean_conda_json(output)

    async def remove_packages(self, env: str, packages: List[str]) -> Dict[str, str]:
        """Delete packages in an environment.

        Args:
            env (str): Environment name
            packages (List[str]): List of packages to delete

        Returns:
            Dict[str, str]: Delete command output.
        """
        ans = await self._execute(
            self.manager, "remove", "-y", "-q", "--json", "-n", env, *packages
        )
        _, output = ans
        return self._clean_conda_json(output)



import json
import os
import random
import shutil
import sys
import tempfile
import unittest
import unittest.mock as mock
import uuid
from itertools import chain
from unittest.mock import AsyncMock

from nb_conda_kernels import CondaKernelSpecManager
from notebook.tests.launchnotebook import assert_http_error

from mamba_gator.envmanager import EnvManager
from mamba_gator.handlers import AVAILABLE_CACHE
from mamba_gator.tests.utils import ServerTest

from .utils import has_mamba


def generate_name() -> str:
    """Generate a random name."""
    return "_" + uuid.uuid4().hex


class JupyterCondaAPITest(ServerTest):
    def setUp(self):
        super(JupyterCondaAPITest, self).setUp()
        self.env_names = []
        self.pkg_name = "astroid"

    def tearDown(self):
        # Remove created environment
        for n in self.env_names:
            self.wait_for_task(self.rm_env, n)
        super(JupyterCondaAPITest, self).tearDown()

    def wait_for_task(self, call, *args, **kwargs):
        r = call(*args, **kwargs)
        self.assertEqual(r.status_code, 202)
        location = r.headers["Location"]
        self.assertRegex(location, r"^/conda/tasks/\d+$")
        return self.wait_task(location)

    def mk_env(self, name=None, packages=None, remove_if_exists=True):
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        new_name = name or generate_name()
        if remove_if_exists and new_name in env_names:
            self.wait_for_task(self.rm_env, new_name)
        self.env_names.append(new_name)

        return self.conda_api.post(
            ["environments"],
            body={"name": new_name, "packages": packages or ["python!=3.10.0"]},
        )

    def rm_env(self, name):
        answer = self.conda_api.delete(["environments", name])
        if name in self.env_names:
            self.env_names.remove(name)
        return answer


class TestChannelsHandler(JupyterCondaAPITest):
    def test_get_list_channels(self):
        answer = self.conda_api.get(["channels"])
        self.assertEqual(answer.status_code, 200)
        data = answer.json()
        self.assertIn("channels", data)
        self.assertIsInstance(data["channels"], dict)

    def test_fail_get(self):
        with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
            error_msg = "Fail to get channels"
            r = {"error": True, "message": error_msg}
            f.return_value = (1, json.dumps(r))
            with assert_http_error(500, msg=error_msg):
                self.conda_api.get(["channels"])

    def test_deployment(self):
        with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
            local_channel = (
                "C:/Users/Public/conda-channel"
                if sys.platform == "win32"
                else "/usr/local/share/conda-channel"
            )
            strip_channel = local_channel.strip("/")
            data = {
                "channel_alias": {
                    "auth": None,
                    "location": "conda.anaconda.org",
                    "name": None,
                    "package_filename": None,
                    "platform": None,
                    "scheme": "https",
                    "token": None,
                },
                "channels": ["defaults", "conda-forge", local_channel],
                "custom_channels": {
                    strip_channel: {
                        "auth": None,
                        "location": "",
                        "name": strip_channel,
                        "package_filename": None,
                        "platform": None,
                        "scheme": "file",
                        "token": None,
                    }
                },
                "custom_multichannels": {
                    "defaults": [
                        {
                            "auth": None,
                            "location": "repo.anaconda.com",
                            "name": "pkgs/main",
                            "package_filename": None,
                            "platform": None,
                            "scheme": "https",
                            "token": None,
                        },
                        {
                            "auth": None,
                            "location": "repo.anaconda.com",
                            "name": "pkgs/r",
                            "package_filename": None,
                            "platform": None,
                            "scheme": "https",
                            "token": None,
                        },
                    ],
                    "local": [
                        {
                            "auth": None,
                            "location": "",
                            "name": strip_channel,
                            "package_filename": None,
                            "platform": None,
                            "scheme": "file",
                            "token": None,
                        }
                    ],
                },
            }
            f.return_value = (0, json.dumps(data))

            response = self.conda_api.get(["channels"])
            self.assertEqual(response.status_code, 200)

            body = response.json()
            channels = body["channels"]
            local = "file:///" if sys.platform == "win32" else "file://"
            expected = {
                strip_channel: [local + local_channel],
                "defaults": [
                    "https://repo.anaconda.com/pkgs/main",
                    "https://repo.anaconda.com/pkgs/r",
                ],
                "conda-forge": ["https://conda.anaconda.org/conda-forge"],
            }
            self.assertEqual(channels, expected)


class TestEnvironmentsHandler(JupyterCondaAPITest):
    def test_get(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)
        envs = self.conda_api.envs()
        env = None
        for e in envs["environments"]:
            if n == e["name"]:
                env = e
                break
        self.assertIsNotNone(env)
        self.assertEqual(env["name"], n)
        self.assertTrue(os.path.isdir(env["dir"]))
        self.assertFalse(env["is_default"])

    def test_failed_get(self):
        with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
            msg = "Fail to get environments"
            err = {"error": True, "message": msg}
            f.return_value = (1, json.dumps(err))
            with assert_http_error(500, msg=msg):
                self.conda_api.get(["environments"])

    def test_root(self):
        self.wait_for_task(self.mk_env, generate_name())
        envs = self.conda_api.envs()
        root = filter(lambda env: env["name"] == "base", envs["environments"])
        self.assertEqual(len(list(root)), 1)

    def test_env_create_and_destroy(self):
        # Creating an existing environment does not induce error
        n = generate_name()
        response = self.wait_for_task(self.mk_env, n)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        self.assertIn(n, env_names)

        response = self.wait_for_task(self.rm_env, n)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        self.assertNotIn(n, env_names)

        response = self.wait_for_task(self.mk_env, n, remove_if_exists=True)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        self.assertIn(n, env_names)

        response = self.wait_for_task(self.rm_env, n)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        self.assertNotIn(n, env_names)

    def test_empty_environment(self):
        n = generate_name()
        self.env_names.append(n)
        response = self.wait_for_task(
            self.conda_api.post, ["environments"], body={"name": n}
        )
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        self.assertIn(n, env_names)

    def cp_env(self, name, new_name):
        self.env_names.append(new_name)
        return self.conda_api.post(
            ["environments"], body={"name": new_name, "twin": name}
        )

    def test_env_clone(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)
        clone_name = n + "-copy"
        response = self.wait_for_task(self.cp_env, n, clone_name)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = map(lambda env: env["name"], envs["environments"])
        self.assertIn(clone_name, env_names)

        response = self.wait_for_task(self.rm_env, clone_name)
        self.assertEqual(response.status_code, 200)
        self.wait_for_task(self.rm_env, n)

    def test_environment_yaml_import(self):
        if has_mamba:
            self.skipTest("FIXME not working with mamba")

        n = generate_name()
        self.env_names.append(n)
        content = """name: test_conda
channels:
- conda-forge
- defaults
dependencies:
- python=3.9
- astroid
prefix: /home/user/.conda/envs/lab_conda
        """

        expected_packages = ["python", "astroid"]

        def g():
            return self.conda_api.post(
                ["environments"],
                body={"name": n, "file": content, "filename": "testenv.yml"},
            )

        response = self.wait_for_task(g)
        self.assertEqual(response.status_code, 200)

        envs = self.conda_api.envs()
        env_names = list(map(lambda env: env["name"], envs["environments"]))
        self.assertIn(n, env_names)

        r = self.conda_api.get(["environments", n])
        pkgs = r.json().get("packages", [])
        package_names = [p["name"] for p in pkgs]
        for p in expected_packages:
            self.assertIn(p, package_names, "{} not found.".format(p))

    def test_environment_text_import(self):
        n = generate_name()
        self.env_names.append(n)
        content = """# This file may be used to create an environment using:
# $ conda create --name <env> --file <this file>
python=3.9
astroid
        """

        expected_packages = ["python", "astroid"]

        def g():
            return self.conda_api.post(
                ["environments"], body={"name": n, "file": content}
            )

        response = self.wait_for_task(g)
        self.assertEqual(response.status_code, 200)

        envs = self.conda_api.envs()
        env_names = list(map(lambda env: env["name"], envs["environments"]))
        self.assertIn(n, env_names)

    def test_update_env_yaml(self):
        if has_mamba:
            self.skipTest("FIXME not working with mamba")

        n = generate_name()
        response = self.wait_for_task(self.mk_env, n, ["python=3.9"])
        self.assertEqual(response.status_code, 200)

        content = """name: test_conda
channels:
- conda-forge
- defaults
dependencies:
- astroid
prefix: /home/user/.conda/envs/lab_conda
        """

        expected_packages = ["astroid"]

        def g():
            return self.conda_api.patch(
                ["environments", n], body={"file": content, "filename": "testenv.yml"},
            )

        response = self.wait_for_task(g)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = list(map(lambda env: env["name"], envs["environments"]))
        self.assertIn(n, env_names)

        r = self.conda_api.get(["environments", n])
        pkgs = r.json().get("packages", [])
        package_names = [p["name"] for p in pkgs]
        for p in expected_packages:
            self.assertIn(p, package_names, "{} not found.".format(p))

    def test_update_env_no_filename(self):
        if has_mamba:
            self.skipTest("FIXME not working with mamba")
            
        n = generate_name()
        response = self.wait_for_task(self.mk_env, n, ["python=3.9"])
        self.assertEqual(response.status_code, 200)

        content = """name: test_conda
channels:
- conda-forge
- defaults
dependencies:
- astroid
prefix: /home/user/.conda/envs/lab_conda
        """

        expected_packages = ["astroid"]

        def g():
            return self.conda_api.patch(["environments", n], body={"file": content},)

        response = self.wait_for_task(g)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = list(map(lambda env: env["name"], envs["environments"]))
        self.assertIn(n, env_names)

        r = self.conda_api.get(["environments", n])
        pkgs = r.json().get("packages", [])
        package_names = [p["name"] for p in pkgs]
        for p in expected_packages:
            self.assertIn(p, package_names, "{} not found.".format(p))

    def test_update_env_txt(self):
        n = generate_name()
        response = self.wait_for_task(self.mk_env, n, ["python=3.9"])
        self.assertEqual(response.status_code, 200)

        content = """# This file may be used to create an environment using:
# $ conda create --name <env> --file <this file>
astroid
        """

        def g():
            return self.conda_api.patch(
                ["environments", n], body={"file": content, "filename": "testenv.txt"},
            )

        response = self.wait_for_task(g)
        self.assertEqual(response.status_code, 200)
        envs = self.conda_api.envs()
        env_names = list(map(lambda env: env["name"], envs["environments"]))
        self.assertIn(n, env_names)

        r = self.conda_api.get(["environments", n])
        self.assertEqual(r.status_code, 200)
        body = r.json()
        self.assertIn("packages", body)
        self.assertGreater(len(body["packages"]), 0)


class TestEnvironmentsHandlerWhiteList(JupyterCondaAPITest):
    @mock.patch("nb_conda_kernels.manager.CACHE_TIMEOUT", 0)
    def test_get_whitelist(self):
        n = "banana"
        self.wait_for_task(self.mk_env, n, packages=["ipykernel",])
        manager = CondaKernelSpecManager()
        manager.whitelist = set(["conda-env-banana-py",])
        env_manager = TestEnvironmentsHandlerWhiteList.notebook.web_app.settings["env_manager"]
        env_manager._kernel_spec_manager = manager

        r = self.conda_api.get(["environments",], params={"whitelist": 1})
        self.assertEqual(r.status_code, 200)
        envs = r.json()
        env = None
        for e in envs["environments"]:
            if n == e["name"]:
                env = e
                break
        self.assertIsNotNone(env)
        self.assertEqual(env["name"], n)
        self.assertTrue(os.path.isdir(env["dir"]))
        self.assertFalse(env["is_default"])
        found_env = len(envs["environments"])
        self.assertEqual(found_env, 2)

        n = generate_name()
        self.wait_for_task(self.mk_env, n)
        r = self.conda_api.get(["environments",], params={"whitelist": 1})
        self.assertEqual(r.status_code, 200)
        envs = r.json()
        self.assertEqual(len(envs["environments"]), found_env)


class TestEnvironmentHandler(JupyterCondaAPITest):
    def test_delete(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)

        r = self.wait_for_task(self.conda_api.delete, ["environments", n])
        self.assertEqual(r.status_code, 200)
        
        # Remove the environment from tracking list since we deleted it manually
        if n in self.env_names:
            self.env_names.remove(n)

        # Remove the environment from tracking list since we deleted it manually
        if n in self.env_names:
            self.env_names.remove(n)

    def test_delete_not_existing(self):
        # Deleting not existing environment does not raise an error
        n = generate_name()

        r = self.wait_for_task(self.conda_api.delete, ["environments", n])
        self.assertEqual(r.status_code, 200)

    def test_get(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)
        r = self.conda_api.get(["environments", n])
        self.assertEqual(r.status_code, 200)
        body = r.json()
        self.assertIn("packages", body)

    def test_fail_get(self):
        n = generate_name()
        with assert_http_error(500, msg="h"):
            self.conda_api.get(["environments", n])

    def test_get_has_update(self):
        n = generate_name()
        self.env_names.append(n)
        self.wait_for_task(
            self.conda_api.post,
            ["environments"],
            body={"name": n, "packages": ["python=3.9", "astroid"]},
        )

        r = self.wait_for_task(
            self.conda_api.get, ["environments", n], params={"status": "has_update"}
        )
        self.assertEqual(r.status_code, 200)
        body = r.json()
        # The response should contain an "updates" key, even if empty
        self.assertIn("updates", body)
        # If updates are available, astroid should be among them
        if len(body["updates"]) > 0:
            names = [p["name"] for p in body["updates"]]
            # Check that we got a valid updates list (could be empty if no updates available)
            self.assertIsInstance(body["updates"], list)

    def test_env_export(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)
        r = self.conda_api.get(["environments", n], params={"download": 1, "history": 0})
        self.assertEqual(r.status_code, 200)

        content = r.text
        self.assertRegex(content, r"name: " + n)
        self.assertRegex(content, r"channels:")
        self.assertRegex(content, r"dependencies:")
        self.assertRegex(content, r"- python=\d\.\d+\.\d+=\w+")
        self.assertRegex(content, r"prefix:")

    def test_env_export_history(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n, packages=["python=3.9"])
        r = self.conda_api.get(
            ["environments", n], params={"download": 1, "history": 1}
        )
        self.assertEqual(r.status_code, 200)

        content = " ".join(r.text.splitlines())
        self.assertRegex(
            content, r"^name:\s" + n + r"\s+channels:(\s+-\s+[^\s]+)+\s+dependencies:\s+-\s+python=3\.9\s+prefix:"
        )

    def test_env_export_not_supporting_history(self):
        try:
            n = generate_name()
            self.wait_for_task(self.mk_env, n)
            EnvManager._conda_version = (4, 6, 0)
            r = self.conda_api.get(
                ["environments", n], params={"download": 1, "history": 1}
            )
            self.assertEqual(r.status_code, 200)

            content = r.text
            self.assertRegex(content, r"name: " + n)
            self.assertRegex(content, r"channels:")
            self.assertRegex(content, r"dependencies:")
            self.assertRegex(content, r"- python=\d\.\d+\.\d+=\w+")
            self.assertRegex(content, r"prefix:")
        finally:
            EnvManager._conda_version = None


class TestCondaVersion(JupyterCondaAPITest):
    def test_version(self):
        EnvManager._conda_version = None
        self.assertIsNone(EnvManager._conda_version)
        self.conda_api.get(
            ["environments",]
        )
        self.assertIsNotNone(EnvManager._conda_version)


class TestPackagesEnvironmentHandler(JupyterCondaAPITest):
    def test_pkg_install_and_remove(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)

        r = self.wait_for_task(
            self.conda_api.post,
            ["environments", n, "packages"],
            body={"packages": [self.pkg_name]},
        )
        self.assertEqual(r.status_code, 200)
        r = self.conda_api.get(["environments", n])
        pkgs = r.json()["packages"]
        v = None
        for p in pkgs:
            if p["name"] == self.pkg_name:
                v = p
                break
        self.assertIsNotNone(v)

        r = self.wait_for_task(
            self.conda_api.delete,
            ["environments", n, "packages"],
            body={"packages": [self.pkg_name]},
        )
        self.assertEqual(r.status_code, 200)
        r = self.conda_api.get(["environments", n])
        pkgs = r.json()["packages"]
        v = None
        for p in pkgs:
            if p["name"] == self.pkg_name:
                v = p
                break
        self.assertIsNone(v)

    def test_pkg_install_with_version_constraints(self):
        test_pkg = "astroid"
        n = generate_name()
        self.wait_for_task(self.mk_env, n, packages=["python=3.9"])

        r = self.wait_for_task(
            self.conda_api.post,
            ["environments", n, "packages"],
            body={"packages": [test_pkg + "==2.14.2"]},
        )
        self.assertEqual(r.status_code, 200)
        r = self.conda_api.get(["environments", n])
        pkgs = r.json()["packages"]
        v = None
        for p in pkgs:
            if p["name"] == test_pkg:
                v = p
                break
        self.assertEqual(v["version"], "2.14.2")

        n = generate_name()
        self.wait_for_task(self.mk_env, n, packages=["python=3.9"])
        r = self.wait_for_task(
            self.conda_api.post,
            ["environments", n, "packages"],
            body={"packages": [test_pkg + ">=2.14.0"]},
        )
        self.assertEqual(r.status_code, 200)
        r = self.conda_api.get(["environments", n])
        pkgs = r.json()["packages"]
        v = None
        for p in pkgs:
            if p["name"] == test_pkg:
                v = tuple(map(int, p["version"].split(".")))
                break
        self.assertGreaterEqual(v, (2, 14, 0))

        n = generate_name()
        self.wait_for_task(self.mk_env, n, packages=["python=3.9"])
        r = self.wait_for_task(
            self.conda_api.post,
            ["environments", n, "packages"],
            body={"packages": [test_pkg + ">=2.14.0,<3.0.0"]},
        )
        self.assertEqual(r.status_code, 200)
        r = self.conda_api.get(["environments", n])
        pkgs = r.json()["packages"]
        v = None
        for p in pkgs:
            if p["name"] == test_pkg:
                v = tuple(map(int, p["version"].split(".")))
                break
        self.assertGreaterEqual(v, (2, 14, 0))
        self.assertLess(v, (3, 0, 0))

    def test_package_install_development_mode(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)

        pkg_name = "banana"
        with tempfile.TemporaryDirectory() as temp_folder:
            os.mkdir(os.path.join(temp_folder, pkg_name))
            with open(os.path.join(temp_folder, pkg_name, "__init__.py"), "w+") as f:
                f.write("")
            with open(os.path.join(temp_folder, "setup.py"), "w+") as f:
                f.write("from setuptools import setup\n")
                f.write("setup(name='{}')\n".format(pkg_name))

            self.wait_for_task(
                self.conda_api.post,
                ["environments", n, "packages"],
                body={"packages": [temp_folder]},
                params={"develop": 1},
            )

            r = self.conda_api.get(["environments", n])
            body = r.json()

            v = None
            for p in body["packages"]:
                if p["name"] == pkg_name:
                    v = p
                    break
            self.assertEqual(v["channel"], "<develop>")
            self.assertEqual(v["platform"], "pypi")

    def test_package_install_development_mode_url_path(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)

        pkg_name = generate_name()[1:]
        folder = os.path.join(self.notebook.contents_manager.root_dir, pkg_name)
        
        try:
            os.mkdir(folder)
            os.mkdir(os.path.join(folder, pkg_name))
            with open(os.path.join(folder, pkg_name, "__init__.py"), "w+") as f:
                f.write("")
            with open(os.path.join(folder, "setup.py"), "w+") as f:
                f.write("from setuptools import setup\n")
                f.write("setup(name='{}')\n".format(pkg_name))

            self.wait_for_task(
                self.conda_api.post,
                ["environments", n, "packages"],
                body={"packages": [pkg_name]},
                params={"develop": 1},
            )

            r = self.conda_api.get(["environments", n])
            body = r.json()

            v = None
            for p in body["packages"]:
                if p["name"] == pkg_name:
                    v = p
                    break
            self.assertEqual(v["channel"], "<develop>")
            self.assertEqual(v["platform"], "pypi")
        finally:
            # Clean up the package directory that was created in the notebook root
            if os.path.exists(folder):
                shutil.rmtree(folder)

    def test_pkg_update(self):
        n = generate_name()
        self.wait_for_task(self.mk_env, n)

        r = self.wait_for_task(self.conda_api.patch, ["environments", n, "packages"])
        self.assertEqual(r.status_code, 200)


class TestPackagesHandler(JupyterCondaAPITest):
    def test_package_search(self):
        r = self.wait_for_task(
            self.conda_api.get, ["packages"], params={"query": self.pkg_name}
        )
        self.assertEqual(r.status_code, 200)
        body = r.json()
        self.assertIn("packages", body)
        v = None
        for p in body["packages"]:
            if p["name"] == self.pkg_name:
                v = p
                break
        self.assertIsNotNone(v)

    def test_package_list_available(self):
        with mock.patch("mamba_gator.handlers.AVAILABLE_CACHE", generate_name()):
            with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
                dummy = {
                    "numpy_sugar": [
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "380115a180acf251faaf754ff37cab8f",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "8bba4c5179a7e40f0c03861df9dfc1fd7827322e76d0b646a29bee055b0b727a",
                            "size": 46560,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "version": "1.0.6",
                        },
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "6306fdf5d1f3fad5049f282b63e95403",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "88e41187218af19e587ef43a3a570a6664d2041cfc01f660eae255a284d9ca77",
                            "size": 46738,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "version": "1.0.8",
                        },
                    ],
                    "numpydoc": [
                        {
                            "arch": None,
                            "build": "py36_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                            "constrains": [],
                            "depends": ["python >=3.6,<3.7.0a0", "sphinx"],
                            "fn": "numpydoc-0.8.0-py36_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "f96891e9071727dfca3ea480408396f6",
                            "name": "numpydoc",
                            "platform": None,
                            "sha256": "8760ab4d1d04b4c9a455baa6961a2885175d74cafac1e06034f99ff7e2357056",
                            "size": 43791,
                            "subdir": "win-64",
                            "timestamp": 1522687759543,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64/numpydoc-0.8.0-py36_0.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_1",
                            "build_number": 1,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.8.0-py_1.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "5e71b7baaecd06f5c2dfbb1055cb0de3",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "30ae298b7e4b02f2e9fe07e1341c70468a95cb0ab6b38dd18d60de7082935494",
                            "size": 21577,
                            "subdir": "noarch",
                            "timestamp": 1531243883293,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.8.0-py_1.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.0-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "081b5590105257246eada8a8bc5dd0aa",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "8ccc9c59c5b874e7f255270e919fd9f8b6e0a4c62dca48bc4de990f5ceb7da34",
                            "size": 32151,
                            "subdir": "noarch",
                            "timestamp": 1555950366084,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch/numpydoc-0.9.0-py_0.tar.bz2",
                            "version": "0.9.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.1-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "de8a98b573872ba539fe7e68e106178a",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "90049dd32972b2e61131ba27c9c4c90b09e701cbba2c6a473d041d7ffa1352c0",
                            "size": 29774,
                            "subdir": "noarch",
                            "timestamp": 1556022967544,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.9.1-py_0.tar.bz2",
                            "version": "0.9.1",
                        },
                    ],
                }
                channels = {
                    "channel_alias": {
                        "auth": None,
                        "location": "conda.anaconda.org",
                        "name": None,
                        "package_filename": None,
                        "platform": None,
                        "scheme": "https",
                        "token": None,
                    },
                    "channels": ["defaults"],
                    "custom_multichannels": {
                        "defaults": [
                            {
                                "auth": None,
                                "location": "repo.anaconda.com",
                                "name": "pkgs/main",
                                "package_filename": None,
                                "platform": None,
                                "scheme": "https",
                                "token": None,
                            }
                        ]
                    },
                    "ssl_verify": False,
                }

                if has_mamba:
                    # Change dummy to match mamba repoquery format
                    dummy = {
                        "result": {
                            "pkgs": list(chain(*dummy.values()))
                        }
                    }

                # Use side_effect to have a different return value for each call
                f.side_effect = [
                    (0, json.dumps(dummy)),
                    (0, json.dumps(channels)),
                ]

                r = self.wait_for_task(self.conda_api.get, ["packages"])
                self.assertEqual(r.status_code, 200)

                args, _ = f.call_args_list[0]
                if has_mamba:
                    self.assertSequenceEqual(args[1:], ["repoquery", "search", "*", "--json"])
                else:
                    self.assertSequenceEqual(args[1:], ["search", "--json"])

                body = r.json()

                expected = {
                    "packages": [
                        {
                            "build_number": [0, 0],
                            "build_string": ["py35_vc14_0", "py35_vc14_0"],
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "name": "numpy_sugar",
                            "platform": None,
                            "version": ["1.0.6", "1.0.8"],
                            "summary": "",
                            "home": "",
                            "keywords": [],
                            "tags": [],
                        },
                        {
                            "build_number": [1, 0, 0],
                            "build_string": ["py_1", "py_0", "py_0"],
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                            "name": "numpydoc",
                            "platform": None,
                            "version": ["0.8.0", "0.9.0", "0.9.1"],
                            "summary": "Sphinx extension to support docstrings in Numpy format",
                            "home": "https://github.com/numpy/numpydoc",
                            "keywords": [],
                            "tags": [],
                        },
                    ],
                    "with_description": True,
                }
                self.assertEqual(body, expected)

    @unittest.skipIf(sys.platform.startswith("win"), "TODO test not enough reliability")
    def test_package_list_available_local_channel(self):
        with mock.patch("mamba_gator.handlers.AVAILABLE_CACHE", generate_name()):
            with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
                dummy = {
                    "numpy_sugar": [
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "380115a180acf251faaf754ff37cab8f",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "8bba4c5179a7e40f0c03861df9dfc1fd7827322e76d0b646a29bee055b0b727a",
                            "size": 46560,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "version": "1.0.6",
                        },
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "6306fdf5d1f3fad5049f282b63e95403",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "88e41187218af19e587ef43a3a570a6664d2041cfc01f660eae255a284d9ca77",
                            "size": 46738,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "version": "1.0.8",
                        },
                    ],
                    "numpydoc": [
                        {
                            "arch": None,
                            "build": "py36_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                            "constrains": [],
                            "depends": ["python >=3.6,<3.7.0a0", "sphinx"],
                            "fn": "numpydoc-0.8.0-py36_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "f96891e9071727dfca3ea480408396f6",
                            "name": "numpydoc",
                            "platform": None,
                            "sha256": "8760ab4d1d04b4c9a455baa6961a2885175d74cafac1e06034f99ff7e2357056",
                            "size": 43791,
                            "subdir": "win-64",
                            "timestamp": 1522687759543,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64/numpydoc-0.8.0-py36_0.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_1",
                            "build_number": 1,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.8.0-py_1.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "5e71b7baaecd06f5c2dfbb1055cb0de3",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "30ae298b7e4b02f2e9fe07e1341c70468a95cb0ab6b38dd18d60de7082935494",
                            "size": 21577,
                            "subdir": "noarch",
                            "timestamp": 1531243883293,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.8.0-py_1.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.0-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "081b5590105257246eada8a8bc5dd0aa",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "8ccc9c59c5b874e7f255270e919fd9f8b6e0a4c62dca48bc4de990f5ceb7da34",
                            "size": 32151,
                            "subdir": "noarch",
                            "timestamp": 1555950366084,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch/numpydoc-0.9.0-py_0.tar.bz2",
                            "version": "0.9.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.1-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "de8a98b573872ba539fe7e68e106178a",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "90049dd32972b2e61131ba27c9c4c90b09e701cbba2c6a473d041d7ffa1352c0",
                            "size": 29774,
                            "subdir": "noarch",
                            "timestamp": 1556022967544,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.9.1-py_0.tar.bz2",
                            "version": "0.9.1",
                        },
                    ],
                }

                if has_mamba:
                    dummy = {
                        "result": {
                            "pkgs": list(chain(*dummy.values()))
                        }
                    }

                with tempfile.TemporaryDirectory() as local_channel:
                    with open(
                        os.path.join(local_channel, "channeldata.json"), "w+"
                    ) as d:
                        d.write(
                            '{ "channeldata_version": 1, "packages": { "numpydoc": { "activate.d": false, "binary_prefix": false, "deactivate.d": false, "description": "Numpy\'s documentation uses several custom extensions to Sphinx. These are shipped in this numpydoc package, in case you want to make use of them in third-party projects.", "dev_url": "https://github.com/numpy/numpydoc", "doc_source_url": "https://github.com/numpy/numpydoc/blob/master/README.rst", "doc_url": "https://pypi.python.org/pypi/numpydoc", "home": "https://github.com/numpy/numpydoc", "icon_hash": null, "icon_url": null, "identifiers": null, "keywords": null, "license": "BSD 3-Clause", "post_link": false, "pre_link": false, "pre_unlink": false, "recipe_origin": null, "run_exports": {}, "source_git_url": null, "source_url": "https://pypi.io/packages/source/n/numpydoc/numpydoc-0.9.1.tar.gz", "subdirs": [ "linux-32", "linux-64", "linux-ppc64le", "noarch", "osx-64", "win-32", "win-64" ], "summary": "Sphinx extension to support docstrings in Numpy format", "tags": null, "text_prefix": false, "timestamp": 1556032044, "version": "0.9.1" } }, "subdirs": [ "noarch" ] }'
                        )
                    local_name = local_channel.strip("/")
                    channels = {
                        "channel_alias": {
                            "auth": None,
                            "location": "conda.anaconda.org",
                            "name": "",  # ← Change None to empty string
                            "package_filename": None,
                            "platform": None,
                            "scheme": "https",
                            "token": None,
                        },
                        "channels": ["defaults"],
                        "custom_channels": {},
                        "custom_multichannels": {
                            "defaults": [
                                {
                                    "auth": None,
                                    "location": "repo.anaconda.com",
                                    "name": "pkgs/main",
                                    "package_filename": None,
                                    "platform": None,
                                    "scheme": "https",
                                    "token": None,
                                }
                            ]
                        },
                    }

                    # Use side_effect to have a different return value for each call
                    f.side_effect = [
                        (0, json.dumps(dummy)),
                        (0, json.dumps(channels)),
                    ]

                    r = self.wait_for_task(self.conda_api.get, ["packages"])
                    self.assertEqual(r.status_code, 200)

                    args, _ = f.call_args_list[0]
                    if has_mamba:
                        self.assertSequenceEqual(args[1:], ["repoquery", "search", "*", "--json"])
                    else:
                        self.assertSequenceEqual(args[1:], ["search", "--json"])

                    body = r.json()

                    expected = {
                        "packages": [
                            {
                                "build_number": [0, 0],
                                "build_string": ["py35_vc14_0", "py35_vc14_0"],
                                "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                                "name": "numpy_sugar",
                                "platform": None,
                                "version": ["1.0.6", "1.0.8"],
                                "summary": "",
                                "home": "",
                                "keywords": [],
                                "tags": [],
                            },
                            {
                                "build_number": [1, 0, 0],
                                "build_string": ["py_1", "py_0", "py_0"],
                                "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                                "name": "numpydoc",
                                "platform": None,
                                "version": ["0.8.0", "0.9.0", "0.9.1"],
                                "summary": "Sphinx extension to support docstrings in Numpy format",
                                "home": "https://github.com/numpy/numpydoc",
                                "keywords": [],
                                "tags": [],
                            },
                        ],
                        "with_description": True,
                    }
                    self.assertEqual(body, expected)

    @unittest.skipIf(sys.platform.startswith("win"), "not reliable on Windows")
    def test_package_list_available_no_description(self):
        with mock.patch("mamba_gator.handlers.AVAILABLE_CACHE", generate_name()):
            with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
                dummy = {
                    "numpy_sugar": [
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "380115a180acf251faaf754ff37cab8f",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "8bba4c5179a7e40f0c03861df9dfc1fd7827322e76d0b646a29bee055b0b727a",
                            "size": 46560,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "version": "1.0.6",
                        },
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "6306fdf5d1f3fad5049f282b63e95403",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "88e41187218af19e587ef43a3a570a6664d2041cfc01f660eae255a284d9ca77",
                            "size": 46738,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "version": "1.0.8",
                        },
                    ],
                    "numpydoc": [
                        {
                            "arch": None,
                            "build": "py36_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                            "constrains": [],
                            "depends": ["python >=3.6,<3.7.0a0", "sphinx"],
                            "fn": "numpydoc-0.8.0-py36_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "f96891e9071727dfca3ea480408396f6",
                            "name": "numpydoc",
                            "platform": None,
                            "sha256": "8760ab4d1d04b4c9a455baa6961a2885175d74cafac1e06034f99ff7e2357056",
                            "size": 43791,
                            "subdir": "win-64",
                            "timestamp": 1522687759543,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64/numpydoc-0.8.0-py36_0.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_1",
                            "build_number": 1,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.8.0-py_1.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "5e71b7baaecd06f5c2dfbb1055cb0de3",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "30ae298b7e4b02f2e9fe07e1341c70468a95cb0ab6b38dd18d60de7082935494",
                            "size": 21577,
                            "subdir": "noarch",
                            "timestamp": 1531243883293,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.8.0-py_1.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.0-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "081b5590105257246eada8a8bc5dd0aa",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "8ccc9c59c5b874e7f255270e919fd9f8b6e0a4c62dca48bc4de990f5ceb7da34",
                            "size": 32151,
                            "subdir": "noarch",
                            "timestamp": 1555950366084,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch/numpydoc-0.9.0-py_0.tar.bz2",
                            "version": "0.9.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.1-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "de8a98b573872ba539fe7e68e106178a",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "90049dd32972b2e61131ba27c9c4c90b09e701cbba2c6a473d041d7ffa1352c0",
                            "size": 29774,
                            "subdir": "noarch",
                            "timestamp": 1556022967544,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.9.1-py_0.tar.bz2",
                            "version": "0.9.1",
                        },
                    ],
                }

                if has_mamba:
                    # Change dummy to match mamba repoquery format
                    dummy = {
                        "result": {
                            "pkgs": list(chain(*dummy.values()))
                        }
                    }

                with tempfile.TemporaryDirectory() as local_channel:
                    local_name = local_channel.strip("/")
                    channels = {
                        "channel_alias": {},
                        "channels": [local_channel],
                        "custom_multichannels": {},
                        "custom_channels": {
                            local_name: {
                                "auth": None,
                                "location": "",
                                "name": local_name,
                                "package_filename": None,
                                "platform": None,
                                "scheme": "file",
                                "token": None,
                            }
                        },
                    }

                    # Use side_effect to have a different return value for each call
                    f.side_effect = [
                        (0, json.dumps(dummy)),
                        (0, json.dumps(channels)),
                    ]

                    r = self.wait_for_task(self.conda_api.get, ["packages"])
                    self.assertEqual(r.status_code, 200)

                    args, _ = f.call_args_list[0]
                    if has_mamba:
                        self.assertSequenceEqual(args[1:], ["repoquery", "search", "*", "--json"])
                    else:
                        self.assertSequenceEqual(args[1:], ["search", "--json"])

                    body = r.json()

                    expected = {
                        "packages": [
                            {
                                "build_number": [0, 0],
                                "build_string": ["py35_vc14_0", "py35_vc14_0"],
                                "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                                "name": "numpy_sugar",
                                "platform": None,
                                "version": ["1.0.6", "1.0.8"],
                                "summary": "",
                                "home": "",
                                "keywords": [],
                                "tags": [],
                            },
                            {
                                "build_number": [1, 0, 0],
                                "build_string": ["py_1", "py_0", "py_0"],
                                "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                                "name": "numpydoc",
                                "platform": None,
                                "version": ["0.8.0", "0.9.0", "0.9.1"],
                                "summary": "",
                                "home": "",
                                "keywords": [],
                                "tags": [],
                            },
                        ],
                        "with_description": False,
                    }
                    self.assertEqual(body, expected)

    def test_package_list_available_caching(self):
        cache_name = generate_name()
        with mock.patch("mamba_gator.handlers.AVAILABLE_CACHE", cache_name):
            with mock.patch("mamba_gator.envmanager.EnvManager._execute", new_callable=AsyncMock) as f:
                dummy = {
                    "numpy_sugar": [
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "380115a180acf251faaf754ff37cab8f",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "8bba4c5179a7e40f0c03861df9dfc1fd7827322e76d0b646a29bee055b0b727a",
                            "size": 46560,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.6-py35_vc14_0.tar.bz2",
                            "version": "1.0.6",
                        },
                        {
                            "arch": None,
                            "build": "py35_vc14_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "constrains": [],
                            "depends": [
                                "cffi",
                                "ncephes",
                                "numba",
                                "numpy",
                                "python 3.5*",
                                "scipy",
                                "vc 14.*",
                            ],
                            "fn": "numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "license": "MIT",
                            "license_family": "MIT",
                            "md5": "6306fdf5d1f3fad5049f282b63e95403",
                            "name": "numpy_sugar",
                            "platform": None,
                            "sha256": "88e41187218af19e587ef43a3a570a6664d2041cfc01f660eae255a284d9ca77",
                            "size": 46738,
                            "subdir": "win-64",
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64/numpy_sugar-1.0.8-py35_vc14_0.tar.bz2",
                            "version": "1.0.8",
                        },
                    ],
                    "numpydoc": [
                        {
                            "arch": None,
                            "build": "py36_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                            "constrains": [],
                            "depends": ["python >=3.6,<3.7.0a0", "sphinx"],
                            "fn": "numpydoc-0.8.0-py36_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "f96891e9071727dfca3ea480408396f6",
                            "name": "numpydoc",
                            "platform": None,
                            "sha256": "8760ab4d1d04b4c9a455baa6961a2885175d74cafac1e06034f99ff7e2357056",
                            "size": 43791,
                            "subdir": "win-64",
                            "timestamp": 1522687759543,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64/numpydoc-0.8.0-py36_0.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_1",
                            "build_number": 1,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.8.0-py_1.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "5e71b7baaecd06f5c2dfbb1055cb0de3",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "30ae298b7e4b02f2e9fe07e1341c70468a95cb0ab6b38dd18d60de7082935494",
                            "size": 21577,
                            "subdir": "noarch",
                            "timestamp": 1531243883293,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.8.0-py_1.tar.bz2",
                            "version": "0.8.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.0-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "081b5590105257246eada8a8bc5dd0aa",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "8ccc9c59c5b874e7f255270e919fd9f8b6e0a4c62dca48bc4de990f5ceb7da34",
                            "size": 32151,
                            "subdir": "noarch",
                            "timestamp": 1555950366084,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy/noarch/numpydoc-0.9.0-py_0.tar.bz2",
                            "version": "0.9.0",
                        },
                        {
                            "arch": None,
                            "build": "py_0",
                            "build_number": 0,
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch",
                            "constrains": [],
                            "depends": ["python", "sphinx"],
                            "fn": "numpydoc-0.9.1-py_0.tar.bz2",
                            "license": "BSD 3-Clause",
                            "md5": "de8a98b573872ba539fe7e68e106178a",
                            "name": "numpydoc",
                            "noarch": "python",
                            "package_type": "noarch_python",
                            "platform": None,
                            "sha256": "90049dd32972b2e61131ba27c9c4c90b09e701cbba2c6a473d041d7ffa1352c0",
                            "size": 29774,
                            "subdir": "noarch",
                            "timestamp": 1556022967544,
                            "url": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/noarch/numpydoc-0.9.1-py_0.tar.bz2",
                            "version": "0.9.1",
                        },
                    ],
                }
                channels = {
                    "channel_alias": {
                        "auth": None,
                        "location": "conda.anaconda.org",
                        "name": None,
                        "package_filename": None,
                        "platform": None,
                        "scheme": "https",
                        "token": None,
                    },
                    "channels": ["defaults"],
                    "custom_multichannels": {
                        "defaults": [
                            {
                                "auth": None,
                                "location": "repo.anaconda.com",
                                "name": "pkgs/main",
                                "package_filename": None,
                                "platform": None,
                                "scheme": "https",
                                "token": None,
                            }
                        ]
                    },
                    "ssl_verify": False,
                }


                if has_mamba:
                    # Change dummy to match mamba repoquery format
                    dummy = {
                        "result": {
                            "pkgs": list(chain(*dummy.values()))
                        }
                    }

                # Use side_effect to have a different return value for each call
                f.side_effect = [
                    (0, json.dumps(dummy)),
                    (0, json.dumps(channels)),
                ]

                # First retrival no cache available
                r = self.wait_for_task(self.conda_api.get, ["packages"])
                self.assertEqual(r.status_code, 200)

                args, _ = f.call_args_list[0]
                if has_mamba:
                    self.assertSequenceEqual(args[1:], ["repoquery", "search", "*", "--json"])
                else:
                    self.assertSequenceEqual(args[1:], ["search", "--json"])

                expected = {
                    "packages": [
                        {
                            "build_number": [0, 0],
                            "build_string": ["py35_vc14_0", "py35_vc14_0"],
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy-forge/win-64",
                            "name": "numpy_sugar",
                            "platform": None,
                            "version": ["1.0.6", "1.0.8"],
                            "summary": "",
                            "home": "",
                            "keywords": [],
                            "tags": [],
                        },
                        {
                            "build_number": [1, 0, 0],
                            "build_string": ["py_1", "py_0", "py_0"],
                            "channel": "https://artifactory.server/artifactory/api/conda/conda-proxy/win-64",
                            "name": "numpydoc",
                            "platform": None,
                            "version": ["0.8.0", "0.9.0", "0.9.1"],
                            "summary": "Sphinx extension to support docstrings in Numpy format",
                            "home": "https://github.com/numpy/numpydoc",
                            "keywords": [],
                            "tags": [],
                        },
                    ],
                    "with_description": True,
                }

                cache_file = os.path.join(tempfile.gettempdir(), cache_name + ".json")
                self.assertTrue(os.path.exists(cache_file))

                with open(cache_file) as cache:
                    self.assertEqual(json.load(cache), expected)

                # Retrieve using cache
                r = self.conda_api.get(["packages"])
                self.assertEqual(r.status_code, 200)
                body = r.json()
                self.assertEqual(body, expected)


class TestTasksHandler(JupyterCondaAPITest):
    def test_get_invalid_task(self):
        with assert_http_error(404):
            self.conda_api.get(["tasks", str(random.randint(1, 1200))])

    def test_delete_invalid_task(self):
        with assert_http_error(404):
            self.conda_api.get(["tasks", str(random.randint(1, 1200))])

    def test_delete_task(self):
        r = self.mk_env()
        location = r.headers["Location"]
        _, index = location.rsplit("/", maxsplit=1)
        r = self.conda_api.delete(["tasks", index])
        self.assertEqual(r.status_code, 204)



import sys

c.ServerProxy.servers = {
    "mamba": {
        "command": [
            sys.executable,
            "-m",
            "mamba_gator",
            "--no-browser",
            '--port={port}',
            "--ip=0.0.0.0",
            "--ServerApp.token=''",
            "--ServerApp.base_url={base_url}mamba",
            "--ServerApp.allow_remote_access=True",
        ],
        "timeout": 120,
        "absolute_url": True,
        "launcher_entry": {
            "enabled": True,
            "icon_path": "/home/jovyan/packages/navigator/style/mamba.svg",
            "title": "Mamba Navigator",
        },
    },
}



from pathlib import Path

import pytest
from packaging.version import Version, InvalidVersion

from mamba_gator.envmanager import EnvManager, parse_version

from .utils import has_mamba


@pytest.mark.skipif(has_mamba, reason="Mamba found")
def test_envmanager_manager_conda():
    manager = EnvManager("", None)
    assert Path(manager.manager).stem == "conda"


@pytest.mark.skipif(not has_mamba, reason="Mamba NOT found")
def test_envmanager_manager_mamba():
    manager = EnvManager("", None)
    assert Path(manager.manager).stem == "mamba"


def test_parse_r_style_version_with_underscore():
    """Test that R-style versions with underscores are converted to dots."""
    result = parse_version("1.8_4")
    assert result == Version("1.8.4")


def test_parse_year_based_version_with_letter():
    """Test that year-based versions like '2023d' are converted correctly."""
    result = parse_version("2023d")
    assert result == Version("2023.4")


def test_parse_invalid_version_returns_none():
    """Test that invalid versions return None instead of raising exceptions."""
    result = parse_version("invalid.version.format!")
    assert result is None


def test_parse_version_with_multiple_underscores():
    """Test that versions with multiple underscores are handled correctly."""
    result = parse_version("1.2_3_4")
    assert result == Version("1.2.3.4")


def test_parse_year_version_with_z_suffix():
    """Test that year versions with 'z' suffix convert to .26."""
    result = parse_version("2023z")
    assert result == Version("2023.26")


def test_parse_standard_semantic_version_unchanged():
    """Test that standard semantic versions pass through unchanged."""
    result = parse_version("1.21.0")
    assert result == Version("1.21.0")


@pytest.mark.asyncio
async def test_list_available_returns_valid_structure():
    """Test that list_available returns the expected data structure."""
    manager = EnvManager("", None)
    result = await manager.list_available()
    assert "packages" in result
    assert "with_description" in result


@pytest.mark.asyncio
async def test_list_available_version_sorting_works():
    """Test that version sorting works correctly after parsing in list_available."""
    manager = EnvManager("", None)
    result = await manager.list_available()

    if result["packages"]:
        # Find a package with multiple versions to test sorting
        multi_version_pkg = None
        for pkg in result["packages"]:
            if isinstance(pkg.get("version"), list) and len(pkg["version"]) > 1:
                multi_version_pkg = pkg
                break

        if multi_version_pkg:
            # Check that versions are in ascending order
            versions = [Version(v) for v in multi_version_pkg["version"]]
            assert versions == sorted(versions)


@pytest.mark.asyncio
async def test_list_available_handles_invalid_versions_gracefully():
    """Test that list_available doesn't crash with invalid versions and logs warnings."""
    manager = EnvManager("", None)

    # This should complete without crashing, even if some packages have invalid versions
    result = await manager.list_available()

    # Should still return valid structure
    assert "packages" in result
    assert isinstance(result["packages"], list)


@pytest.mark.asyncio
async def test_list_available_preserves_package_metadata():
    """Test that version parsing doesn't break other package metadata."""
    manager = EnvManager("", None)
    result = await manager.list_available()

    if result["packages"]:
        # Check that packages still have their expected metadata
        package = result["packages"][0]
        expected_fields = {"name", "version", "channel"}
        assert all(field in package for field in expected_fields)

        # Version should be a list of strings
        assert isinstance(package["version"], list)
        assert all(isinstance(v, str) for v in package["version"])


@pytest.mark.asyncio
async def test_list_available_version_parsing_integration():
    """Test the full integration of version parsing within list_available."""
    manager = EnvManager("", None)
    result = await manager.list_available()

    if result["packages"]:
        # Look for packages that might have had version parsing applied
        for package in result["packages"]:
            if package["version"]:
                # All versions should be valid strings after parsing
                for version_str in package["version"]:
                    assert isinstance(version_str, str)
                    assert len(version_str) > 0

                    # Should be parseable by packaging.Version after our parsing
                    try:
                        Version(version_str)
                    except InvalidVersion:
                        # If this fails, our version parsing didn't work correctly
                        pytest.fail(f"Version '{version_str}' from package '{package['name']}' is not valid after parsing")


@pytest.mark.asyncio
async def test_list_available_build_metadata_preserved():
    """Test that build numbers and strings are correctly aligned with parsed versions."""
    manager = EnvManager("", None)
    result = await manager.list_available()

    if result["packages"]:
        for package in result["packages"]:
            versions = package.get("version", [])
            build_numbers = package.get("build_number", [])
            build_strings = package.get("build_string", [])

            if versions and build_numbers:
                # All arrays should have the same length
                assert len(versions) == len(build_numbers)

            if versions and build_strings:
                assert len(versions) == len(build_strings)



import json
import os.path as osp

# flake8: noqa
try:
    from ._version import __version__
except ImportError:
    # Fallback when using the package in dev mode without installing
    # in editable mode with pip. It is highly recommended to install
    # the package from a stable release or in editable mode: https://pip.pypa.io/en/stable/topics/local-project-installs/#editable-installs
    import warnings

    warnings.warn("Importing 'mamba_gator' outside a proper installation.")
    __version__ = "dev"

from .handlers import _load_jupyter_server_extension
from .navigator.main import MambaNavigator


# Jupyter Extension points
def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": "@mamba-org/gator-lab"}]


def _jupyter_server_extension_points():
    return [{"module": "mamba_gator"}, {"module": "mamba_gator.navigator.main", "app": MambaNavigator}]


# For backward compatibility
load_jupyter_server_extension = _load_jupyter_server_extension



import datetime
import errno
import json
import os
import sys
import time
from binascii import hexlify
from subprocess import CalledProcessError, check_call
from tempfile import TemporaryDirectory
from threading import Event, Thread
from typing import List
from unittest.mock import patch

import jupyter_core.paths
from notebook.tests.launchnotebook import NotebookTestBase as ServerTestBase
from tornado.ioloop import IOLoop
from traitlets.config import Config

from mamba_gator.handlers import NS

# TODO: Remove this compatibility layer when migrating to JupyterLab 4
# JupyterLab 4+ and Jupyter Notebook 7+ both use jupyter_server exclusively
try:
    from jupyter_server.auth import IdentityProvider
    from jupyter_server.serverapp import ServerApp
    from jupyter_server.utils import url_path_join
    USE_JUPYTER_SERVER = True
except ImportError:
    # Legacy fallback for Jupyter Notebook 6.x - can be removed with JupyterLab 4 migration
    from notebook.notebookapp import NotebookApp as ServerApp
    from notebook.utils import url_path_join
    USE_JUPYTER_SERVER = False



TIMEOUT = 150
SLEEP = 1

try:
    check_call(["mamba", "--version"])
except (CalledProcessError, FileNotFoundError):
    has_mamba = False
else:
    has_mamba = True


class APITester(object):
    """Wrapper for REST API requests"""

    url = "/"

    def __init__(self, request):
        self.request = request

    def _req(self, verb: str, path: List[str], body=None, params=None):
        if body is not None:
            body = json.dumps(body)
        response = self.request(
            verb, url_path_join(self.url, *path), data=body, params=params
        )

        if 400 <= response.status_code < 600:
            try:
                response.reason = response.json()["message"]
            except Exception:
                pass
        response.raise_for_status()

        return response


class JupyterCondaAPI(APITester):
    """Wrapper for nbconvert API calls."""

    url = NS

    def delete(self, path: List[str], body=None, params=None):
        return self._req("DELETE", path, body, params)

    def get(self, path: List[str], body=None, params=None):
        return self._req("GET", path, body, params)

    def patch(self, path: List[str], body=None, params=None):
        return self._req("PATCH", path, body, params)

    def post(self, path: List[str], body=None, params=None):
        return self._req("POST", path, body, params)

    def envs(self):
        return self.get(["environments"]).json()


class ServerTest(ServerTestBase):
    # Force extension enabling - Disabled by parent class otherwise
    # TODO: Remove NotebookApp config when migrating to JupyterLab 4
    config = Config({
        "NotebookApp": {"nbserver_extensions": {"mamba_gator": True}},
        "ServerApp": {"jpserver_extensions": {"mamba_gator": True}}
    })

    @classmethod
    def setup_class(cls):
        # Copy paste from https://github.com/jupyter/notebook/blob/6.0.0/notebook/tests/launchnotebook.py
        #  Only to suppress setting PYTHONPATH with sys.path
        #   For notebook v6 we could overwrite get_env_patch but unfortunately it is not available for Python 3.5
        cls.tmp_dir = TemporaryDirectory()

        def tmp(*parts):
            path = os.path.join(cls.tmp_dir.name, *parts)
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
            return path

        cls.home_dir = tmp("home")
        data_dir = cls.data_dir = tmp("data")
        config_dir = cls.config_dir = tmp("config")
        runtime_dir = cls.runtime_dir = tmp("runtime")
        cls.notebook_dir = tmp("notebooks")
        cls.env_patch = patch.dict(
            "os.environ",
            {
                "HOME": cls.home_dir,
                "IPYTHONDIR": os.path.join(cls.home_dir, ".ipython"),
                "JUPYTER_NO_CONFIG": "1",  # needed in the future
                "JUPYTER_CONFIG_DIR": cls.config_dir,
                "JUPYTER_DATA_DIR": cls.data_dir,
                "JUPYTER_RUNTIME_DIR": cls.runtime_dir,
            },
        )
        cls.env_patch.start()
        cls.path_patch = patch.multiple(
            jupyter_core.paths,
            SYSTEM_JUPYTER_PATH=[tmp("share", "jupyter")],
            ENV_JUPYTER_PATH=[tmp("env", "share", "jupyter")],
            SYSTEM_CONFIG_PATH=[tmp("etc", "jupyter")],
            ENV_CONFIG_PATH=[tmp("env", "etc", "jupyter")],
        )
        cls.path_patch.start()

        config = cls.config or Config()
        config.NotebookNotary.db_file = ":memory:"

        cls.token = hexlify(os.urandom(4)).decode("ascii")
        
        # TODO: Simplify this when migrating to JupyterLab 4 - only keep the jupyter_server branch
        if USE_JUPYTER_SERVER:
            config.ServerApp.identity_provider_class = IdentityProvider
            config.IdentityProvider.token = cls.token
            config.ServerApp.allow_unauthenticated_access = True
        else:
            config.NotebookApp.token = cls.token

        started = Event()

        def start_thread():
            if "asyncio" in sys.modules:
                import asyncio

                asyncio.set_event_loop(asyncio.new_event_loop())
            app = cls.notebook = ServerApp(
                port=cls.port,
                port_retries=0,
                open_browser=False,
                config_dir=cls.config_dir,
                data_dir=cls.data_dir,
                runtime_dir=cls.runtime_dir,
                notebook_dir=cls.notebook_dir,
                base_url=cls.url_prefix,
                config=config,
                allow_root=True,
                token=cls.token,
            )
            # don't register signal handler during tests
            app.init_signal = lambda: None
            # clear log handlers and propagate to root for nose to capture it
            # needs to be redone after initialize, which reconfigures logging
            app.log.propagate = True
            app.log.handlers = []
            app.initialize()
            app.log.propagate = True
            app.log.handlers = []
            loop = IOLoop.current()
            loop.add_callback(started.set)
            try:
                app.start()
            finally:
                # set the event, so failure to start doesn't cause a hang
                started.set()
                app.session_manager.close()

        cls.notebook_thread = Thread(target=start_thread)
        cls.notebook_thread.daemon = True
        cls.notebook_thread.start()
        started.wait()
        cls.wait_until_alive()

    def setUp(self):
        super(ServerTest, self).setUp()
        self.conda_api = JupyterCondaAPI(self.request)

    def wait_task(self, endpoint: str):
        start_time = datetime.datetime.now()
        if endpoint.startswith("/" + NS):
            endpoint = endpoint[len(NS) + 1 :]

        while (datetime.datetime.now() - start_time).total_seconds() < TIMEOUT:
            time.sleep(SLEEP)
            response = self.conda_api.get([endpoint])
            response.raise_for_status()
            if response.status_code != 202:
                return response

        raise RuntimeError("Request {} timed out.".format(endpoint))

# Disable Windows file association dialogs during testing
# This prevents Windows from showing "How do you want to open this file?" dialogs
# when conda operations encounter files with unknown extensions during testing
if sys.platform == "win32":
    os.environ["PATHEXT"] = ""



import os
import pytest

from mamba_gator.envmanager import RUNNER_COMMAND, get_env_path


@pytest.mark.parametrize(
    "spec,expected",
    [
        (
            {
                "argv": [
                    *RUNNER_COMMAND,
                    "/path/to/conda",
                    os.path.join("path", "to", "envs", "myenv"),
                    "-m",
                    "ipykernel_launcher",
                ],
                "metadata": {},
            },
            os.path.join("path", "to", "envs", "myenv"),
        ),
        (
            {
                "argv": [
                    *RUNNER_COMMAND,
                    "/path/to/conda",
                    os.path.join("path", "to", "envs", "myenv"),
                    "-m",
                    "ipykernel_launcher",
                ],
                "metadata": {
                    "conda_env_path": os.path.join("path", "to", "envs", "myenv")
                },
            },
            os.path.join("path", "to", "envs", "myenv"),
        ),
        (
            {
                "argv": [os.path.join("path", "to", "envs", "myenv", "bin", "python")],
                "metadata": {},
            },
            os.path.join("path", "to", "envs", "myenv"),
        ),
        ({"argv": [], "metadata": {}}, None),
    ],
)
def test_get_env_path(spec, expected):
    assert get_env_path(spec) == expected
